"""Provides core "Spectra" class and associated utilities."""

import string
import cPickle
import logging
from types import NoneType, MethodType
from operator import itemgetter
import datetime

import numpy as np
from scipy import integrate

from pandas import DataFrame, DatetimeIndex, Index, Series, read_csv, MultiIndex
from pandas.core.common import _is_bool_indexer
from pandas.core.indexing import _is_list_like, _is_nested_tuple

# skspec imports 
from skspec.core.specindex import SpecIndex
from skspec.core.abcindex import ConversionIndex, CustomIndex
from skspec.core.specstack import SpecStack
from skspec.core.abcspectra import ABCSpectra, SpecError

import skspec.core.utilities as pvutils
import skspec.config as pvconfig
from skspec.units.abcunits import IUnit, Unit

# Merge
from skspec.pandas_utils.metadframe import MetaDataFrame, MetaSeries
from skspec.logger import decode_lvl, logclass
from skspec.plotting import PLOTPARSER, _genplot, _gen2d3d, areaplot
from skspec.exceptions import badkey_check, badcount_error, RefError, BaselineError

from matplotlib.dates import date2num as _d2num


logger = logging.getLogger(__name__) 


### _normdic, from_T, to_T and spec_slice all used to be in speclabeltool.py###
_normdic={None:'No Normalization', #Don't change without updating normplot; relies on these keys 
      't':'Transmittance', 
      '%t':'(%)Transmittance', 
      'r':'Inverse Transmittance (1/T)',
      'a':'Absorbance (base 10)', 
      'a(ln)':'Absorbance (base e)'} 


### Remember, data is already divided by ref before from_T is called, so "r" is base unit, not "t".  
from_T={'t':lambda x: 1.0/x, 
        '%t':lambda x: 100.0 * (1.0/ x), 
        'r':lambda x: x, 
        'a':lambda x: -np.log10(x),
        'a(ln)':lambda x:-np.log(x)}


to_T={'t':lambda x: 1.0/x, 
      '%t':lambda x: (1.0/ x) / 100.0, 
      'r':lambda x: x, 
      'a':lambda x: np.power(10, -x), 
      'a(ln)':lambda x: np.exp(-x)} 


def spec_slice(spectral_array, bins):
   """ Simple method that will divide a spectral index into n evenly sliced
   bins and return as nested tuples. Useful for generating wavelength slices 
   with even spacing.  Simply a wrapper around np.histogram.
   """
   edges=np.histogram(spectral_array, bins)[1]
   return [ (edges[idx], edges[i]) for idx, i in enumerate( range(1, len(edges)))]

# Unit validations
def _valid_xunit(value, dic):
   """ Validates existence of key (usually a unit type like spectral unit in a dictionary such as specunits)"""
   if value == None:
      return None
   else:
      badkey_check(value, dic.keys())
      return value.lower()    


def _valid_norm(sout):
   """When user is switching spectral intensity units, make sure they do it write."""
   return _valid_xunit(sout, _normdic)

# Used by varunits, specunits, etc... if user doesn't realize their api
# doesn't support UNITS.  Don't delete
def _valid_indextype(index):
   """ Checks index object to make sure it subclasses from ConversionIndex.
   Used for methods like as_specunit(), where only a valid operation if
   underlying Index is a SpecIndex.  Custom Spectra that use, say a Float64Index
   would not support such unit system operations.  Probably a cornercase;
   most skspec Spectra subclasses will strictly type Index.
   """
   try:
      index.unitshortdict
   except AttributeError:
      raise SpecError('Unit API only supported by skspec ConversionIndex'
                      ' got type %s' % type(index))    


def specplot(ts, *args, **pltkwargs):   
   """ Plotting interface.  Use kind='surf, wire etc...' to go through 
   various plot types.  Will append correct x and y labels.
   
   Parameters
   ----------
   
   norm : None or str
       Shortcut to return timespectra in normalized ref system (ie Absorbance)
       without changing the Spectra.
       
   """
   
   kind = pltkwargs.pop('kind', 'spec')   
   norm = pltkwargs.pop('norm', None)

   # FORCE DEFAULT COLOR/COLORMAPS
   if 'cmap' not in pltkwargs and 'color' not in pltkwargs and 'colors' not in pltkwargs:
      if kind == 'spec':
         pltkwargs['cmap'] = pvconfig.CMAP_1DSPECPLOT
      elif kind == 'contour':
         pltkwargs['cmap'] = pvconfig.CMAP_CONTOUR
      elif PLOTPARSER.is_3d(kind):        
         pltkwargs['color'] = pvconfig.COLOR_3DPLOT
      
   
   if norm:
      if ts.norm != norm:  #Better/general way to do this? (_normdic.keys())
         ts = ts.as_norm(norm)      

   if not ts._base_sub:
      logger.warn('Spectrum does not have subtracted baseline; could affect '
                           'result in specious absorbance data.')          

   # Area plot gets special axis label   
   if kind == 'area':
      pltkwargs.setdefault('xlabel', ts.full_varunit)
   else:
      pltkwargs.setdefault('xlabel', ts.full_specunit)

   pltkwargs.setdefault('title', ts.name)            

   pltfcn, is_2d_3d = PLOTPARSER[kind].function, PLOTPARSER.is_2d_3d(kind)
   
   if is_2d_3d:
      pltkwargs.setdefault('ylabel', ts.full_varunit)      
      pltkwargs.setdefault('zlabel', ts.full_iunit) 
   
   else:
      pltkwargs.setdefault('ylabel', ts.full_iunit)
               
   # Hack 
   if pltfcn == _gen2d3d:
      pltkwargs['kind'] = kind
      
   ax = pltfcn(ts, *args, **pltkwargs)

   # Some plot methods return (ax, mappable), some just ax
   try:
      ax = ax[0]
   except TypeError:
      pass
   
   # Reverse X-axis for all PLOT TYPES (EXCEPT CONTOUR)
   if ts.index[0] > ts.index[-1]:
      if kind != 'contour':
         if kind == 'waterfall':
            ax.set_ylim(ax.get_ylim()[::-1]) # HACK
         else:
            ax.set_xlim(ax.get_xlim()[::-1]) 

   return ax


########################################
## Spectra Public Utilities    #####
######################################## 


# Wrapper for df_from_directory
def spec_from_dir(directory, csvargs, sortnames=False, concat_axis=1, shortname=True, cut_extension=False):
   """ Takes files from a directory, presuming they are identically formatted, and reads them into
   a dataframe by iterating over read_csv().  All arguments that read_csv() would take are passed
   in.  These kwargs are in regard the files themselves, for example, skiprows, skipfooter and header.

   For now, no support for glob to take only files of a certain file extension.
   For now, conctaentation occurs along

   Args:
      directory- Path to directory where files are stored.
      csvargs- Dictionary of arguments that are passed directly to the read_csv() function. For example
               skiprows, header, na_values etc... see pandas API for read_csv()

   Kwds:
      sortnames- Will attempt to autosort the filelist. Otherwise, files are ordered by the module
                 os.path.listdir().

      concat_axis- How to merge datafiles into a dataframe.  Default is axis=1, which means all files 
                   should share the same index values.  I use this for spectral data where the wavelength
                   column is my dataframe index.

      shortname- If false, full file path is used as the column name.  If true, only the filename is used. 

      cut_extension- If kwd shortname is True, this will determine if the file extension is saved or cut from the data."""

   return Spectra(df_from_directory(directory, 
                                    csvargs,
                                    sortnames=sortnames, 
                                    concat_axis=concat_axis, 
                                    shortname=shortname, 
                                    cut_extension=cut_extension))


class Spectrum(ABCSpectra, MetaSeries):
   """ Compliment to pandas Series: single array of spectral values with
   spectral Index.
   """
   
   def __init__(self, *args, **kwargs):
      
      specifier = kwargs.pop('specifier', pvconfig.SPECIFIERDEF)
      # CUSTOMIZE!?
      super(Spectrum, self).__init__(*args, **kwargs)        
      self.specifier = specifier
        

   @property
   def unit(self):
      return self.index._unit.short   
   
   @property
   def full_unit(self):
      return self.index._unit.full

   @property
   def _header(self):
      """ Header for string printout """
      delim = pvconfig.HEADERDELIM
      
      # Certain methods aren't copying this correctly.  Delete later if fix

      unit = pvutils.safe_lookup(self, 'full_unit')
      iunit = pvutils.safe_lookup(self, 'full_iunit')
      
   
      header = "*%s*%s [%s]%sIunit: %s\n" % \
            (self.name, 
             delim, 
             unit,
             delim, 
             iunit)
      
      return header


   @property
   def _header_html(self):
      """ Header for html printout """
      
      delim = pvconfig.HEADERHTMLDELIM
   
      if self.ndim > 1:
         colorshape = '<font color="#0000CD">(%s X %s)</font>' % (self.shape)
      else:
         colorshape = '<font color="#0000CD"> (%s)</font>' % (self.shape)

      unit = pvutils.safe_lookup(self, 'full_unit')
      iunit = pvutils.safe_lookup(self, 'full_iunit')

      header = "%s&nbsp%s%s Index Unit:&nbsp%s%sIunit:&nbsp%s" % \
         (self.name, 
          colorshape,
          delim,
          unit,
          delim,
          iunit)   
      
      return header   
   
   def plot(self, *args, **pltkwds):

      # Replace w/ set defaults
      pltkwds.setdefault('linewidth', 3.0)
      pltkwds.setdefault('color', 'black') # If removing, colormap default in _genplot
                                           # Will cause bug
      
      pltkwds.setdefault('xlabel', self.unit)
      pltkwds.setdefault('ylabel', self.specifier)
      pltkwds.setdefault('title', '%s' % self.name)
      
      return _genplot(self, *args, **pltkwds)
   
   @classmethod
   def from_series(cls, spectra, series, **kwargs):
      """ When Spectra methods need to return a Spectrum, this is called
      so that metadata is properly transferred.
      """

      # Will I ever want to construct from something other than Series?
      if not isinstance(series, Series):
         raise SpecError('Spectrum.from_series() requires series input,'
                         'got %s' % type(series))
     
      out = cls(series.values, index=series.index)

      def _transfer(attr):
         '''If attr not in kwargs, trasnfer from spectra object'''

         # Should work for all valid attr, since spectra sets its own defaults 
         if attr in kwargs:
            value = kwargs[attr]
         else:
            value = pvutils.hasgetattr(spectra, attr, 'invalid transfer')
         setattr(out, attr, value)
      
      for attr in ['baseline', 
#                   'varunit', 
#                   'full_varunit',
                   'name', 
                   'norm', 
                   'full_norm', 
                   '_iunit'    #Pass private variable, because iunit defined in ABC spec
                   ]:            #Otherwise, properties like specunit become just attributes
                           
         _transfer(attr)
      
      return out
   
   
   
class Spectra(ABCSpectra, MetaDataFrame):
   """ Provides core Spectra composite pandas DataFrame to represent a set 
   of spectral data.  Enforces spectral data along the index and temporal 
   data as columns.  The spectral index is controlled from the specindex module, 
   which has a psuedo-class called SpecIndex (really a monkey patched Index). 
   Temporal data is stored using a DatetimeIndex or a modified interval 
   reprentation.  The need to support two types of temporal index, one of 
   which is Pandas builtin DatetimeIndex is what led me to not create a 
   special index object (like SpecIndex).  Other spectral axis types (like 
   Temperature axis) should probably be built close to the manner of Spec index.
   The Spectra dataframe actually stores enough temporal metadatato go back 
   and forth between DatetimeIndex and Interval representations.  
   It does this by generating one or the other on the fly, and never relies on 
   the current label object to generate teh next object.
   """

   def __init__(self, *dfargs, **dfkwargs):

      self._strict_index = dfkwargs.pop('strict_index', SpecIndex)
      self._strict_columns = dfkwargs.pop('strict_columns', None)

      # Pop default DataFrame keywords before initializing###
      self.name = str(dfkwargs.pop('name', ''))

      # Spectral index-related keywords
      specunit = dfkwargs.pop('specunit', None)
      varunit = dfkwargs.pop('varunit', Unit())
      iunit = dfkwargs.pop('iunit', IUnit(short='cts', 
                                         full='Counts (photons)',
                                         symbol=r'$\gamma$') # ever used? 
                                         )
      
      # Intensity data-related stuff
      norm = dfkwargs.pop('norm', None)

      # Time index-related keywords  (note, the are only used if a 
      # DatetimeIndex is not passed in)
      reference = dfkwargs.pop('reference', None)        
      bline = dfkwargs.pop('baseline', None)

      # Logging __init__() in @logclass can't access self.name
      logger.info('Initializing %s' %  '%s:(name = %s)' % 
                  (self.__class__.__name__, self.name))


      super(Spectra, self).__init__(*dfargs, **dfkwargs)        

      # Iunit
      self.iunit = iunit

      # Convert to the passed unit        
      self._frame.index = self._valid_index(self.index)
      if specunit:
         self.specunit = specunit
#         self._frame.index = self._frame.index.convert(specunit)          

      # Convert to the passed unit
      self._frame.columns = self._valid_columns(self.columns)            
      if varunit:
         self.varunit = varunit
#         self._frame.columns = self._frame.columns.convert(varunit)     

      # Assign spectral intensity related stuff but 
      # DONT CALL _set_normtype function
      norm =_valid_norm(norm)
      self._normtype = norm

      # This has to be done AFTER self._frame has been set
      self._reference = self._reference_valid(reference)#SHOULD DEFAULT TO NONE SO USER CAN PASS NORMALIZED DATA WITHOUT REF        

      # Baseline Initialization (UNTESTED)
      self._base_sub = False 
      self._baseline = None  

      if not isinstance(bline, NoneType):
         self.baseline = bline

      # Store intrinsic attributes for output later by listattr methods
      self._intrinsic=self.__dict__.keys()
      self._intrinsic.remove('name') #Not a private attr

      # Which attributes/methods are manipulated along with the dataframe
      self._cnsvdattr=['_reference', '_baseline']#, 'strict_index', 'strict_columns']
      self._cnsvdmeth=['_slice', 'pvutils.boxcar'] #_slice is ix


   def _comment(self, statement, level='info'):
      prefix = ''
      if self.name:
         prefix = self.name + ': '

      statement = '%s%s' % (prefix, statement)
      logger.log(decode_lvl(level), statement)


   ###########################################################################    
   # Methods ( @property decorators ensure use can't overwrite methods) ####   
   #############################################################################

   # Unit printouts (change to pretty print?)
   def _list_out(self, outdic, delim='\t'):
      """ Generic output method for shortname:longname iterables.  Prints out various
      dictionaries, and is independent of the various datastructures contained"""
      print '\nKey',delim,'Description'
      print '-------------------\n'

      for (k,v) in sorted(outdic.items()):
         print k,delim,v
      print '\n'

   def varunits(self, delim='\t', pprint = False):
      """ Print out all available variable units in a nice format"""
      _valid_indextype(self.columns)
      out = self.columns.unitshortdict
      if not pprint:
         return out
      return self._list_out(out, delim=delim)        

   def norms(self, delim='\t', pprint = False):
      """ Intensity units of dateframe.  Eg %Transmittance vs. Absorbance"""
      if not pprint:
         return _normdic
      return self._list_out(_normdic, delim=delim)

   @property
   def plot_kinds(self):
      """ List available plot kinds"""
      print PLOTPARSER.__shortrepr__()

   # Properties, private, etc...?
   def _summary(self):
      """"""
      
      
   def _summary_html(self):
      """"""""

   # Self necessary here or additional df stuff gets printed   
   def specunits(self, delim='\t', pprint = False):
      """ Print out all available units in a nice format"""
      _valid_indextype(self.index)
      out = self.index.unitshortdict
      if not pprint:
         return out
      return self._list_out(out, delim=delim) 

   # ADD PYTHON STRING FORMATTING TO THIS
   def list_attr(self, privattr=False, dfattr=False, methods=False, types=False, delim='\t', sortby='name'): #values=False,):
      """ Outputs various attributes in the namespace, as well as their types.

      Parameters
      ----------
         privattr: Should private attributes be listed.
         dfattr: Should dataframe's attributes (eg shape, columns) be listed
         methods: If True, methods will be output; otherwise, only non-method attributes are listed.
         types: Output the type of attribute along with it?
         delim: For formatting.
         sortby: If outputing by type, order by name or by type?

      Notes
      -----
         Would like to add "values" but the formatting might get messy.  A good IDE can provide this information anyway,
         for example, Stack Data in windware.  Also increases complexity of the logic below.
         Refer to label_stats() for more detailed information on certain attributes.
      ."""

      # Take all attributes in current instance        
      atts=[x for x in dir(self) if '__' not in x]

      # Remove self._intrinsic (aka class attributes) if desired
      if privattr==False:
         atts=[attr for attr in atts if attr not in self._intrinsic]        
         filterout=['_intrinsic', 'ix', '_ix', 'list_attr'] #Don't want these as attributes
         for att in filterout:
            atts.remove(att)

      # Include dataframe attributes if desired
      if dfattr==True:
         atts=atts+[x for x in dir(self._frame) if '__' not in x]

      # Remove methods if desired
      if methods==False:
         atts=[attr for attr in atts if isinstance(getattr(self, attr), MethodType) != True]
      else:
         atts.append('ix') #This as an attribute but want it to seem like a method

      # Should output include types and/or values?
      outheader=['Attribute']
      if types==True: 
         outheader=outheader+['Type']
         atts=[(att, str(type(getattr(self, att))).split()[1].split('>')[0] ) for att in atts]
         #string operation just does str(<type 'int'>) ---> 'int'


         # Sort output either by name or type
         badkey_check(sortby, ['name', 'type']) #sortby must be 'name' or 'type'

         if sortby == 'name':
            atts=sorted(atts, key=itemgetter(0))
         elif sortby == 'type':
            atts=sorted(atts, key=itemgetter(1))

      # Output to screen
      print delim.join(outheader)    
      print '--------------------'
      for att in atts:
         if types == True:
            print string.rjust(delim.join(att), 7) #MAKE '\N' comprehension if string format never works out
         else:
            print att    


   ########################
   # ts.refrence operations
   ########################

   @property
   def reference(self):
      """ This is stored as a Series unless user has set it otherwise."""
      if self._reference is not None:
         return Spectrum(self._reference.values, self._reference.index)

   @reference.setter
   def reference(self, reference, force_series=True):  
      """ Before changing reference, first validates.  Then considers various cases, and changes 
      class attributes and dataframe values appropriately."""

      int_or_column = False
      if reference in self._frame.columns or isinstance(reference, int):            
         int_or_column = True

      # Adding or changing reference
      if reference is not NoneType:

         # If data is in raw/full mode (normtype=None)
         if self._normtype == None:
            reference=self._reference_valid(reference, force_series=force_series)                  

            self._reference=reference

         # Let _set_normtype() do lifting.  Basically convert to full and back to current normtype. 
         else:
            self._set_normtype(self._normtype, ref=reference)
            if int_or_column:            
               logger.warn('Reference changed based on column data while'
                           'Spectra was normalized.')            

      # Removing reference.  
      else:
         # If current reference is not None, convert
         if not isinstance(self._reference, NoneType):  #Can't do if array==None
            self._set_normtype(None, ref=self._reference)
            self._reference=None


   def _reference_valid(self, ref, force_series=True, zero_warn=True, nan_warn=True):
      """ Helper method for to handles various scenarios of valid references.  Eg user wants to 
      convert the spectral representation, this evaluates manually passed ref vs. internally stored one.

      Parameters:
      -----------
      ref: reference object for inspection.   See notes for more details of validation process.
      force_series:  If true, reference is forced to a Series type.
      zero_warn: If true, a warning is raised if reference contains zeros.
      nan_warn:  If true, a warning is raised if reference contains nans.

      Notes:
      ------

      This goes through several cases to determine reference.  In order, it attempts to get the reference from:
          1. Column name from dataframe.
          2. Column index from dataframe.
          3. Uses ref itself.
             a. If series, checks for compatibility between spectral indicies.
             b. If dataframe, ensures proper index and shape (1-d) and converts to series (if force_series True).
             c. If iterable, converts to series (if force_series True)

      Errors will be thrown if new reference does not have identical spectral values to self._frame.index as evaluated by
      numpy.np.array_equal()

      Zero_warn and nan_warn are useful when the user plans on changing intensity units of the data.  For example,
      converting from raw data to absorbance data, a relative division will result in hard-to-spot errors due to
      the 0's and nans in the original reference. 
      """


      if isinstance(ref, NoneType):
         return ref

      # First, try ref is itself a column name
      if ref in self._frame.columns:
         rout=self._frame[ref]

      # If rtemp is an integer, return that column value.  
      # NOTE: IF COLUMN NAMES ARE ALSO INTEGERS, THIS CAN BE PROBLEMATIC.
      elif isinstance(ref, int):
         rout=self._frame[self._frame.columns[ref]]        

      # Finally if ref is itself a series, make sure it has the correct spectral index
      elif isinstance(ref, Series):
         if np.array_equal(ref.index, self._frame.index) == False:
            raise RefError(ref.index, self)
         else:
            rout=ref

      # Finally if array or other iterable, force to a series
      else:
         if force_series:

            # If user passes dataframe, downconvert to series 
            if isinstance(ref, DataFrame):
               if ref.shape[1] != 1:
                  raise TypeError('reference must be a dataframe of a single column with index values equivalent to those of %s'%self._name)
               if ref.index.all() != self._frame.index.all():
                  raise RefError(ref, self)                    
               else:
                  rout=Series(ref[ref.columns[0]], index=ref.index) 

            # Add index to current iterable 
            else:
               if len(ref) != len(self._frame.index):
                  raise RefError(ref, self)
               else:
                  rout=Series(ref, index=self._frame.index)

         # Return itrable as is
         else:
            rout=ref

      if nan_warn:
         if np.isnan(np.sum(rout)):  #Fast way to check for nan's
            logger.error('Nans found in reference.  Error raised because '
                         '"nan_warn" = True in _reference_valid()')


      if zero_warn:
         if 0.0 in rout:
            logger.error('Zero values found in reference.  Error raised '
                         'because "zero_warn" = True in _reference valid()')

      return rout  #MAKES MORE SENSE TO MAKE THIS A 1D DATAFRAME


   def apply(self, *args, **kwargs):
      """ Raw=True seems to produce correct results more often from some
      #deep issue that is not fully tracked. """
      kwargs.setdefault('raw', True)
      specifier = kwargs.pop('specifier', args[0].__name__)
      # Call _dataframe attribute and transfer attributes
      out = self._framegetattr('apply', *args, **kwargs)
      
      ## If Spectrum, use function name as specifier!
      if isinstance(out, Spectrum):
         out.specifier = specifier
      return out
   
   def wavelength_slices(self, ranges, apply_fcn='mean', **applyfcn_kwds):
      """Returns sliced averages/sums of wavelength regions. Composite dataframe will nicely
      plot when piped into spec aesthetics timeplot.


      Parameters:
      -----------
      apply_fcn: Chooses the way to collapse data.  Builtin functions include 'mean'
                'sum', 'simps', 'trapz', 'romb', 'cumtrapz', but any user function that
                 results in collapsed data (eg averging function/integrators/etc...) can 
                 be passed in with any relevant keywords.

      **apply_fcn_kdws: 
                 If user is passing a function to apply_fcn that requires keywords, 
                 the get passed in to dfcut.apply() 

      Notes:
      -----_
      See description of 'df_wavelength_slices' in utilities.py for more information.         
      For easy plotting, plot the transpose of the returned timespectra.  
      """   

      dflist=[]; snames=[]

      if isinstance(ranges, float) or isinstance(ranges, int):
         ranges=spec_slice(self.index, ranges)           

      # If single range is passed in, want to make sure it can still be iterated over...
      if len(ranges)==2:
         ranges=[ranges]

      for rng in ranges:
         if len(rng)!=2:
            raise AttributeError("In slices function, all ranges passed in must be len 2, aka a start and stop \
                pair.  %s of length %s was entered" % (rng, len(rng)))
         else:
            dfcut=self.ix[rng[0]:rng[1]]
            snames.append('%s:%s'%(rng[0],rng[1]))

            if isinstance(apply_fcn, str):

               # For inverted xaxis
               if self.index[0] > self.index[-1]:
                  xvals = self.index[::-1]
               else:
                  xvals = self.index
               # Pandas cython operations ###
               if apply_fcn.lower() == 'mean': 
                  series=dfcut.mean(axis=0)
               elif apply_fcn.lower() == 'sum':
                  series=dfcut.sum(axis=0)


               # Integration isn't the same as sum because it takes the units of the x-axis into account through x
               # parameter.  If interval is odd, last interval is used w/ trapezoidal rule
               elif apply_fcn.lower() == 'simps':
                  series=dfcut.apply(integrate.simps, x=xvals, even='last')

               elif apply_fcn.lower() == 'trapz':
                  series=dfcut.apply(integrate.trapz, x=xvals)     

               elif apply_fcn.lower() == 'romb':
                  series=dfcut.apply(integrate.romb, x=xvals)

               elif apply_fcn.lower() == 'cumtrapz':
                  series=dfcut.apply(integrate.trapz, x=xvals, ititial=None)           

               else:
                  raise AttributeError('apply_fcn in wavelength slices, as a string, must be one of the following: \
                        (mean, sum, simps, trapz, romb, cumtrapz) but %s was entered.  Alternatively, you can pass \
                        a function to apply_fcn, with any relevant **kwds')

            # Try to apply arbirtrary function.  Note, function can't depend on dfcut, since 
            # that is created in here
            else:
               series=dfcut.apply(apply_fcn, **applyfcn_kwds)


            # OK
            dflist.append(series)

      # Out is a series (e.g. Area)
      if len(dflist) == 1:
         return dflist[0]
      return self._transfer(DataFrame(dflist, index=snames))


   def boxcar(self, binwidth, axis=1):
      """Performs pvutils.boxcar averaging by binning data.

      Parameters
      ----------
      binwidth: int or float
         Width of the bin.  EG 10minutes or 5 seconds.
         
      axis: 0 or 1
         Average over rows or columns, respectively.
      """
      if axis == 0 and self.varunit == 'dti':
         raise NotImplementedError('Cannot pvutils.boxcar along DateTime axis.')

      return self._transfer(pvutils.boxcar(self, binwidth=binwidth, axis=axis))


   def area(self, apply_fcn='simps'):
      """ Returns total area under the spectra vs. time curve.  To choose a slice of the spectrum,
          call wavelength_slices with appropriate ranges and an apply_fcn appropriate to style of integration.

      Parameters
      ----------
         
      apply_fcn: Integration method-'simps', 'trapz', 'romb', 'cumtrapz'
         

      Notes
      -----
         This is a wrapper for wavelength_sclices, and as such, user can actually pass any function into the apply_fcn. 
         Besides mean/sum which is understands already, one could, for example, pass a euler integration function and
         it should work. 

         WHEN PLOTTING, PLOT THE TRANSPOSE OF THE RETURNED DF.
      """

      out = self.wavelength_slices((self.index[0],#min(self.index),
                                     self.index[-1]),#max(self.index)), 
                                    apply_fcn=apply_fcn)

      out.specifier = 'Area (%s)' % apply_fcn
      return out


   # Spectral column attributes/properties
   ### SPECUNIT IS JUST CARRIED THROUGH ON DF._INDEX.  ANYCHANGES WILL
   ### RETURN A NEW INDEX, AND ALSO UPDATE SPECUNIT IN SAID INDEX!
   @property
   def specunit(self):
      return self._frame.index._unit.short    #Short name key

   @property
   def full_specunit(self):
      return self._frame.index._unit.full

   @specunit.setter
   def specunit(self, unit):
      # Calling self.index will type-check      
      if isinstance(unit, Unit):
         self.index = CustomIndex(self.index, unit=unit)
      
      else:
         _valid_indextype(self.index) #Raises error in corner cases (leave)               
         self.index = self._frame.index.convert(unit) 
         
         
   @property
   def varunit(self):
      return self._frame.columns._unit.short    #Short name key

   @property
   def full_varunit(self):
      return self._frame.columns._unit.full

   @varunit.setter
   def varunit(self, unit):
      # Calling self.columns will type-check
      if isinstance(unit, Unit):
         self.columns = CustomIndex(self.columns, unit=unit) #self.columns.setter will validate
      
      # If string
      else:
         _valid_indextype(self.columns) #Raises error in corner cases (leave)      
         self.columns = self._frame.columns.convert(unit) 
               
      

   def as_specunit(self, unit):
      """ Returns new dataframe with different spectral unit on the index."""
      tsout = self.deepcopy()
      tsout.specunit = unit
      return tsout       

   def as_varunit(self, unit):
      """ Returns new dataframe with different spectral unit on the index."""
      tsout = self.deepcopy()
      tsout.varunit = unit
      return tsout   

   # Still want this?
   def set_specindex(self, start=None, stop=None, spacing=None, unit=None):
      """
      Helper method to generate a spectral index.  Works similarly to
      pandas.daterange().  Takes in start, stop and spacing.  Users can use
      either start and stop, start and spacing or stop and spacing (but not 
      start, stop and spacing) to generate specindex.

      Parameters
      ----------
      start: Wavelength start value.
      stop: Wavelength stop value.
      spacing: Wavelength spacing.
      unit: Valid specunit type.  Note, if None is passed, current specunit of
            object is used.  To force a null unit, use special keyword 'null'.

      Notes
      -----
      User must pass 2 of the 3 keywords start, stop or spacing.  Function
      will compute the other using the length of the current dfindex.  
      For example, if df.index is 100 entries and user enters start=0, stop=1000,
      then spacing is set to ten to retain 100 entries.  
      If user enters start and spacing, then stop is determined to retain 100
      entries.
      User can also enter stop and spacing, and start is inferred by counting
      backwards from stop at steps equal to spacing size.

      To generate less restricted SpecIndex from start, stop AND spacing, try
      set_sindex() function in specindex.py module.

      Under the hood this calls set_sindex() which is a wrapper to
      np.linspace(). 
      """        

      numpts=float(len(self.index))

      # If no unit passed, conserve current one
      if not unit:
         unit=self.specunit

      # If user wants to force a null unit
      else:
         if unit.lower()=='null':
            unit=None

      # Bad keyword cases
      if start and stop and spacing:
         raise badcount_error(2,3,3, argnames='start, stop, keywords')

      if not start and not stop and not spacing:
         raise badcount_error(2,0,3, argnames='start, stop, keywords')

      # User enters start and stop    
      if start and stop:
         start=float(start)            
         stop=float(stop)

      # User enters spacing and start or spacing and stop.
      if spacing:    
         spacing=float(spacing)            
         if start:
            start=float(start)                
            stop= start + (spacing * numpts)           
         elif stop:
            stop=float(stop)
            start=stop - (spacing * numpts)

      # If user only entered one keyword, then all three will not be generated by this point 
      if not start or not stop:
         raise badcount_error(2,1,3, argnames='start, stop, keywords')

      self._frame.index = self._strict_index(np.linspace(start, stop, numpts), unit=unit)


   ###################################
   # Baseline related operations ###
   ###################################

   # SHOULD ALSO ENFORCE DATATYPES
   def _base_gate(self):
      """ Quick check to see if self._baseline is found or missing.  If missing,
          raises error.  If found, passes."""
      if isinstance(self._baseline, NoneType):
         raise AttributeError('Baseline not found.')        

   def sub_base(self):
      """ Subtracts baseline from entire dataset.

          Notes:
          -----
            Does have to call self._frame.  Just doing self.sub will not work, even though
            calling ts.sub() from an outside program does in fact work.  Behavior is due
            to way in which python classes deal with overwrites of self."""

      # self._baseline is not none
      self._base_gate()

      # Only subtract if baseline isn't currently subtracted
      if not self._base_sub:
         # Index, although should be correct, is type object and is getting falses for entries...
         logger.critical('Subtracting baseline, but may not have all: elements being equal.  Fix index')
         self._frame = self._frame.sub(self._baseline, axis=0)
         if self._reference is not None:
            self._reference = self._reference.sub(self._baseline, axis=0)
         self._base_sub = True

      else:
         logger.warn('Baseline is already subtracted.')

   def add_base(self):
      """ Adds baseline to data that currently has it subtracted."""

      # self._baseline is not none
      self._base_gate()

      # Only add if baseline is currently in a subtracted state
      if self._base_sub:
         self._frame = self._frame.add(self._baseline, axis=0)   
         if self._reference is not None:
            self._reference = self._reference.add(self._baseline, axis=0)
            self._base_sub = False

      #else:
         #print 'raise waring? already added'        

   @property
   def base_sub(self):
      """ Is baseline currently subtracted from spectral data."""
      self._base_gate()
      return self._base_sub


   def _valid_baseline(self, baseline):
      """ Validates user-supplied baseline before setting.  Mostly,
          ensures that baseline.index == self._frame.index"""


      if baseline is None:
         self._comment('Baseline is None')
         return 

      # If baseline not iterable, return series of constant values
      try:
         baseline.__iter__
      except AttributeError:
         self._comment('Baseline is not iterable; filling with constant '
                       'value %s' % baseline)
         return Series(baseline, index=self._frame.index)


      # If Series compare index values
      # NEED A SPEC INDEX CLASS, THEN REMOE MOST OF THIS
      if isinstance(baseline, Series):
         if baseline.shape != self.index.shape:
            raise BaselineError('Baseline shape mismatch %s vs %s %s'
                                %(baseline.shape, self.name, self.index.shape))
         else:
            if np.array_equal(baseline.index, self._frame.index):
               self._comment('Baseline is a series; has identical index ' 
                             'to self._frame.index')
               return baseline
            else:
               self._comment('Baseline is a series; shape identical to'
                             ' self.index.shape; but elements are not equal.')
               mytype = self._frame.index.dtype
               bline_type = baseline.index.dtype

               try:
                  baseline.index.dtype = mytype
               except Exception:
                  raise BaselineError('Baseline index (%s) could not be cast to  '
                                      'self.index.type (%s)' % (bline_type, mytype))
               else:
                  logger.warn('Baseline index type (%s) converted to self.index '
                              'type (%s)' % (bline_type, mytype))
                  if np.array_equal(baseline.index, self._frame.index):
                     return baseline
                  else:
                     raise BaselineError('Baseline index != self._frame.index')

   # If other type of iterable, make series
      else:
         # Make sure length is correct
         if len(baseline) == len(self._frame.index):
            self._comment("Baseline is iterable, but not a Series; converting")
            return Series(baseline, index=self._frame.index, dtype=float)
         else:
            raise BaselineError('Baseline must be of length %s to '
                                'match the current spectral index.'%len(self._frame.index))


   def plot(self, *args, **pltkwargs):   
      """ Plotting interface.  Use kind='surf, wire etc...' to go through 
      various plot types.  Will append correct x and y labels.
      """      
      return specplot(self, *args, **pltkwargs)
            

   @property
   def baseline(self):
      # Should these be stored as specrum?  Cause right now, just converting on return
      if self._baseline is not None:
         return Spectrum(self._baseline.values, index=self._baseline.index)


   @baseline.setter
   def baseline(self, bline):
      """ Allows user to set a baseline curve with a variety of options."""
      ## Logic.  Need validation for various cases.
      # Depending on type/length, force or set index?  What about non-equal length types,
      # try an interpolation?
      # End th "valid_base" 

      bline=self._valid_baseline(bline)

      # Data does not current contain subtracted baseline
      if not self._base_sub:
         self._baseline = bline

      # Data does contain subtracted baseline
      else:
         self.add_base()
         self._baseline=bline
         # If baseline is not none, go ahead and re-subtract
         if bline is not None:
            self.sub_base()


   # Spectral Intensity related attributes/conversions
   @property
   def full_norm(self):
      return _normdic[self._normtype]

   @property
   def norm(self):
      return self._normtype

   @norm.setter
   def norm(self, unit):     
      """ Change norm in place."""
      if isinstance(unit, basestring):
         if unit.lower() in ['none', 'full']:
            unit=None

      self._set_normtype(unit)        


   def as_norm(self, unit, reference=None):
      """ Returns new Spectra of modified norm.  Useful if in-place operation not desirable.  
      Also has the option of manually passing a new reference for on-the-fly rereferencing."""
      if isinstance(unit, basestring):
         if unit.lower() in ['none', 'full']:
            unit=None


      tsout = self.deepcopy()        
      tsout._set_normtype(unit, reference)
      return tsout


   def _set_normtype(self, sout, ref=None):
      """Function used to change spectral intensity representation in a convertible manner. Not called on
      initilization of Spectra(); rather, only called by as_norm() method."""

      sout = _valid_norm(sout)
      sin = self._normtype
      df = self._frame #Could also work just by calling self...

      # Corner case, for compatibility with reference.setter
      if sin==None and sout==None:
         return

      ########################################################################
      # Case 1: User converting from full data down to referenced data.#####
      ########################################################################
      if sin==None and sout != None:
         rout=self._reference_valid(ref)

         # If user tries to downconvert but doesn't pass reference, use stored one
         if rout == None:
            rout = self._reference

         # If ref not passed, use current reference.  Want to make sure it is 
         # not none, but have to roundabout truthtest
         if isinstance(rout, NoneType):
            raise TypeError('Cannot convert spectrum to norm %s without a reference'%sout)                
         else:
            df=pvutils.divby(df, divisor=rout)
            df=df.apply(from_T[sout])                    
         #Assumes typeerror is due to rout not being a series/array, but doesn't further test it.  EG do another
         #typeerror check of try: if rout == None: raise error


      ##############################################################   
      # Case 2: Changing spectral representation of converted data.#
      ##############################################################     
      elif sin !=None and sout != None:

         # If user changing reference on the fly, need to change ref ###
         if not isinstance(ref, NoneType): #and ref != self._reference:
            rout=self._reference_valid(ref)

            # Make this it's own method called change reference?
            df=df.apply(to_T[sin])
            df=df.mul(self._reference, axis=0)  
            df=pvutils.divby(df, divisor=rout)
            df=df.apply(from_T[sout])                        


         else:
            rout=self._reference #For sake of consistent transferring at end of this function
            df=df.apply(to_T[sin])
            df=df.apply(from_T[sout])        


      ###########################################################    
      # Case 3: User converting referenced data up to full data.#
      #############################################################
      elif sin !=None and sout==None:
         rout=self._reference_valid(ref)

         if rout == None:
            rout=self._reference

         df=df.apply(to_T[sin])
         df=df.mul(rout, axis=0)  #Multiply up!

      self._reference=rout       
      self._normtype=sout        
      self._frame=df    

   ############################################ 
   #####Overwrite MetaDataFrame behavior ########
   ############################################    


   def _valid_index(self, index):
      """ Recast index to self._strict_index """
      if self._strict_index:       
         if not isinstance(index, self._strict_index):

            # Cornercase User passes PressureIndex where SpecIndex required
            if isinstance(index, ConversionIndex) or isinstance(index, CustomIndex):
               raise SpecError('Invalid index type: %s; require %s.' % 
                               (type(index), self._strict_index))                
            try:
               index = self._strict_index(index)
            except Exception:
               raise SpecError('Could not convert index to %s.' % 
                               self._strict_index )
            else:
               # If subclassing ConversionIndex
               try:
                  unit = self._frame.index.unit
               except AttributeError:
                  pass
               else:
                  index = index.convert(unit)                        

      return index


   def _valid_columns(self, columns):
      """ Recast columns to self._strict_columns """
      if self._strict_columns:        
         if not isinstance(columns, self._strict_columns):

            # User passes a valid skspec index type    
            if isinstance(columns, ConversionIndex) or isinstance(columns, CustomIndex):
               raise SpecError('Invalid column type: %s; require %s.' % 
                               (type(columns), self._strict_columns))


            # Attempts to force type conversion into self._strict_column
            try:
               columns = self._strict_columns(columns)
            except Exception:
               raise SpecError('Could not convert columns to %s.' % 
                               self._strict_columns )
            try:                    
               # If subclassing ConversionIndex
               unit = self._frame.columns.unit
            except AttributeError:
               pass
            else:
               columns = columns.convert(unit)                                 

      return columns


   # Properies
   # ---------------
 

   # Necessary even though these are already defined in metadataframe
   # Setter looks for them on this object
   @property
   def index(self):
      return self._frame.index

   @property
   def columns(self):
      return self._frame.columns


   @index.setter
   def index(self, index):
      self._frame.index = self._valid_index(index)    


   @columns.setter
   def columns(self, columns):
      self._frame.columns = self._valid_columns(columns)


   @property
   def cnsvdattr(self):
      """ attr:value of conserved attributes"""
      csvdout={}
      for attr in self._cnsvdattr:
         if hasattr(self, attr):
            csvdout[attr]=getattr(self, attr)
         else:
            csvdout[attr]=None
      return csvdout

   @cnsvdattr.setter
   def cnsvdattr(self, attrs):
      """ Set which attributes are mutated in DataFrame operations via cons_methods."""
      if isinstance(attrs, basestring):
         attrs=[attrs]

      consout=[]
      for attr in attrs:
         try:
            consout.append(attrgetter(attr, self))
         except AttributeError:
            raise AttributeError('%s not found on object %s'%(attr, self)) 

      self._cnsvdattr=consout

   @property
   def cnsvdmeth(self):
      return self._cnsvdmeth

   @cnsvdmeth.setter
   def cnsvdmeth(self, attrs):
      # Do I even want to do anything more than setter?
      if isinstance(attrs, basestring):
         attrs=[attrs]

      for attr in attrs:
         if hasattr(self, attr):
            pass
         else:
            raise AttributeError('%s has no attribute "%s"'%(self.name, attr))

      self._cnsvdmeth=attrs        

   def _framegetattr(self, attr, *fcnargs, **fcnkwargs):
      """ Overloading MEtaPandas Object, returns a new object, either
      Spectra or Spectrum with attributes properly conserved.  Handles special
      cases, for example, if returning Spectrum instead of Spectra.  Will
      conserve attributes such as baseline/reference.

      Notes:
      -----
      Since new series items always have a name by default, an originally unnamed series
      attribute (such as reference) will end with a name=_reference afterwards.  Added a hack
      to intercept this highly common attribute and apply to output."""

      out=getattr(self._frame, attr)(*fcnargs, **fcnkwargs)

      # If operation returns a dataframe, return new Spectra
      if isinstance(out, DataFrame):

         # Are there specially conserved attributes?
         csvdout=None
         if attr in self._cnsvdmeth:    
            if self._cnsvdattr:
               cnsvdattr=dict((k,v) for k, v in self.cnsvdattr.iteritems() 
                              if v is not None)

               csvdf=DataFrame(cnsvdattr)  #STILL WORKS WITH NONEQUAL LENGTH
               _csvdfattrs=dict((attr, (cnsvdattr[attr].__dict__)) for attr in cnsvdattr)

               try:
                  csvdout=getattr(csvdf, attr)(*fcnargs, **fcnkwargs)
               except Exception:
                  raise Exception('Could not successfully perform operation "%s" on one or multiple \
                        conserved attributes %s.'%attr, '","'.join(csvdout.columns))                               

         # Create new timespectra object
         tsout = self._transfer(out)

         # Apply conserved attributes for example, if baseline was sliced
         if not isinstance(csvdout, NoneType):

            for col in csvdout:

               # Restore custom attributes on the cnsvdattributes
               restattr=[attr for attr in _csvdfattrs[col] if attr not in 
                         csvdout[col].__dict__]
               if restattr:
                  for attr in restattr:
                     setattr(csvdout[col], attr, _csvdfattrs[col][attr]) 

                  # Hack to conserve "name" attribute of series return
                  try:
                     setattr(csvdout[col], 'name', _csvdfattrs[col]['name']) 
                  except KeyError:
                     pass

               # Apply conserved attributes to new dataframe                        
               setattr(tsout, col, csvdout[col])

         return tsout

      # Otherwise return whatever the method return would be
      elif isinstance(out, Series):
         return Spectrum.from_series(self, out)
         
      else:
         return out
      
   @property
   def _header(self):
      
      delim = pvconfig.HEADERDELIM
      
      header = super(Spectra, self)._header
      
      if self._baseline is None:
         base = 'None'
      else:
         if self._base_sub:
            base = 'Subtracted'
         else:
            base = 'Found (no sub)'
            
      if self.reference is None:
         ref = 'None'
      else:
         ref = 'Found'
         

      header += '\nBaseline: %s%sReference: %s%sNormalization: %s\n' %\
                                                               (base, 
                                                                delim, 
                                                                ref,
                                                                delim, 
                                                                str(self.norm))
      return header
   
      
   @property
   def _header_html(self):
      """ Add normalization, baseline, reference. """
      delim = pvconfig.HEADERHTMLDELIM
      
      #http://www.w3schools.com/html/html_colors.asp
      def _red(string):
         return '<font color="#CC0033">%s</font>' % (string) 
      
      def _green(string):
         return '<font color="#009900">%s</font>' % (string) 
         
      
      header = super(Spectra, self)._header_html
      
      if self.norm is None:
         norm = _red('None')
      else:
         norm = _green(self.full_norm) #Actually use the value
         
      if self._baseline is None:
         base = _red('None')
      else:
         if self._base_sub:
            base = _green('Subtracted')
         else:
            base = _green('Found')+' (no sub)'
         
      if self.reference is None:
         ref = _red('None')
      else:
         ref = _green('Found')
         

      header += '<br><br>Baseline: %s%sReference: %s%sNormalization: %s' %(base, 
                                                              delim, 
                                                              ref,
                                                              delim, 
                                                              norm)
      return header
   

   def __getitem__(self, keyslice):
      ''' Item lookup.  If output is an interable, _transfer is called.  
      Sometimes __getitem__ returns a float (indexing a series) at which 
      point we just want to return that.'''

      dfout = self._frame.__getitem__(keyslice)
       
      if isinstance(dfout, Series):
         return Spectrum.from_series(self, dfout)
       
      return super(Spectra, self).__getitem__(keyslice)
    


   def split_by(self, n, axis=1, stack=True, **stackkwargs):
      """ Slice data into n subsets.

      Notes
      -----
      If n does not evenly divide the axis, the final element returned tends
      to be larger than the others (culmination of rounding down iteratively)
      and a warning is raised.
      """
      ts_split = pvutils.split_by(self, n, axis=1, astype=list)
      stackkwargs.setdefault('name', self.name)
      if stack:
         ts_split = SpecStack(ts_split, **stackkwargs)
      return ts_split


   # Output
   #-------
   
   def to_R_df(self):
      """ Wraps pandas.rpy.common import convert_to_r_dataframe."""

      from pandas.rpy.common import convert_to_r_dataframe
      logger.info('Converting %s to R dataframe.' % self.full_name)
      return( convert_to_r_dataframe(self._frame) )      

   def to_csv(self, path_or_buff, meta_separate=None, **csv_kwargs):
      """ Output to CSV file.  

          Parameters:
          ----------
             path_or_buff: string path to outfile destination.

             meta_separate: 
                 If None: metadata is lost and self._frame.to_csv(**csv_kwargs) is called.
                 If False: metadata is serialized and output at tail file the path_or_buff file.
                 If True: metadata is added to it's own file named path_or_buff.mdf

             csv_kwargs: csv formatting arguments passed directoy to self._frame.to_csv()

          Notes:
          ------
             MetaData is gotten from self.__dict__.
             In future, may opt to implement the option to choose the meta_separate filename if 
             output is separated (eg meta_separate=True)

      """

      meta=cPickle.dumps(self.__dict__)
      self._frame.to_csv(path_or_buff, **csv_kwargs)
      if meta_separate == None:
         return
      elif meta_separate == True:
         raise NotImplemented('Not yet implemented this style of csv output')
      elif meta_separate==False:
         logger.critical('Adding metadata to csv; feature is UNDER CONSTRUCTION...')
         o=open(path_or_buff, 'a') #'w'?#
         o.write(meta)
         o.close()


   # CLASS METHODS
   # -------------


   @classmethod
   def from_csv(cls, filepath_or_buffer, header_datetime=False, 
                index_datetime=False, **kwargs):
      """ Read from CSV file.  Wrapping pandas read_csv:
      http://pandas.pydata.org/pandas-docs/version/0.13.1/  \
      generated/pandas.io.parsers.read_csv.html

      Parameters:
      ----------
      filepath_or_buffer: (from pandas API) string or file handle / StringIO. 
      The string could be a URL. Valid URL schemes include http, ftp, s3, 
      and file. For file URLs, a host is expected. For instance, a local 
      file could be file.

      header_datetime: bool or format str (False)
          Used to parse files with header that is a datetime string.  
          User must pass a format string e.g. %Y/%m/%d %H:%M:%S, which
          is the default.  Results in DatetimeIndex as label.  This will
          set the varunit to 'dti' automatically, unless specifically
          set as None.

      index_datetime: bool or format str (False)
          Used to parse files with header that is a datetime string.  
          User must pass a format string e.g. %Y/%m/%d %H:%M:%S, which
          is the default.   Results in DatetimeIndex as label.  This will
          set the specunit to 'dti' automatically, unles specifically set
          as None.

      **kwargs: Any valid spectra or pandas readcsv() kwargs.       

      Returns: Spectra
      """


      _CSVKWDS = dict(sep=',', 
                      dialect=None, 
                      compression=None,
                      doublequote=True, 
                      escapechar=None,
                      quotechar='"', 
                      quoting=0, 
                      skipinitialspace=False, 
                      lineterminator=None,
                      header='infer', 
                      index_col=None, 
                      names=None, 
                      prefix=None,
                      skiprows=None, 
                      skipfooter=None, 
                      skip_footer=0, 
                      na_values=None, 
                      na_fvalues=None, 
                      true_values=None, 
                      false_values=None,
                      delimiter=None, 
                      converters=None, 
                      dtype=None, 
                      usecols=None, 
                      engine=None, 
                      delim_whitespace=False,
                      as_recarray=False,
                      na_filter=True, 
                      compact_ints=False, 
                      use_unsigned=False, 
                      low_memory=True,
                      buffer_lines=None, 
                      warn_bad_lines=True, 
                      error_bad_lines=True,
                      keep_default_na=True,
                      thousands=None,
                      comment=None, 
                      decimal='.',
                      parse_dates=False, 
                      keep_date_col=False,
                      dayfirst=False, 
                      date_parser=None, 
                      memory_map=False, 
                      nrows=None,
                      iterator=False,
                      chunksize=None, 
                      verbose=False,
                      encoding=None, 
                      squeeze=False, 
                      mangle_dupe_cols=True, 
                      tupleize_cols=False,
                      infer_datetime_format=False)

      # Pop out any parser keyword args
      for kw in kwargs.keys():
         if kw in _CSVKWDS:
            _CSVKWDS[kw] = kwargs[kw]
            del kwargs[kw]

      df = read_csv(filepath_or_buffer, **_CSVKWDS)

      if header_datetime:
         if not isinstance(header_datetime, basestring):
            header_datetime = '%Y/%m/%d %H:%M:%S'      

         df.columns = DatetimeIndex([datetime.datetime.strptime(s, header_datetime) 
                                     for s in df.columns])

         if 'varunit' in kwargs:
            if kwargs['varunit'] != None and kwargs['varunit'] != 'dti':
               raise SpecError("from_csv() option 'header_dateime' requires"
                               "varunit = 'dti' or 'None', received %s" % kwargs['varunit'])

         else:
            kwargs['varunit'] = 'dti'   

      if index_datetime:
         if not isinstance(index_datetime, basestring):
            index_datetime = '%Y/%m/%d %H:%M:%S'      

         df.index = DatetimeIndex([datetime.datetime.strptime(s, index_datetime) 
                                   for s in df.index])

         if 'specunit' in kwargs:
            if kwargs['specunit'] != None and kwargs['specunit'] != 'dti':
               raise SpecError("from_csv() option 'index_dateime' requires"
                               "specunit = 'dti' or 'None', received %s" % kwargs['specunit'])

         else:
            kwargs['specunit'] = 'dti' 
      return cls(df, **kwargs) 


   @classmethod
   def from_series(cls, pandas_object, **dfkwargs):
      """ Return a Spectra from a similiar object, either a pandas DataFrame
      or a skspec Spectra subclass.  Useful for type-casting, for example
      have a DataFrame and want it as a TimeSpectra, which will do its
      own index conversions.
      """
      return cls(np.array(pandas_object.df), 
                 index=pandas_object.index, 
                 columns=pandas_object.columns,
                 **dfkwargs)

    

## TESTING ###
if __name__ == '__main__':  

   from skspec.data import solvent_evap, aunps_glass, trip_peaks
   import matplotlib.pyplot as plt
   ts = aunps_glass()
   ts.baseline=5
   print 'hi'
   print ts.baseline
   x = ts.iloc[0, 0:5].index
   print x
   print ts.mean().index

 


''' Plotting wrappers for dataframes, for implementation in skspec.  Basically calls df.plot()
with some extra bells and whistles like trying to extract column labels and dataframe names,
as well as some exploration into custom labels and color mapping.'''

__author__ = "Adam Hughes"
__copyright__ = "Copyright 2012, GWU Physics"
__license__ = "Free BSD"
__maintainer__ = "Adam Hughes"
__email__ = "hugadams@gwmail.gwu.edu"
__status__ = "Development"

import numpy as np

import matplotlib.colors as mplcolors
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import skspec.config as cnfg

from skspec.exceptions import badvalue_error

DEFCOLOR = (1.0, 0.0, 0.0)

_rgb_from_string = mplcolors.ColorConverter().to_rgb

class ColorError(Exception):
    """ """

def cmget(color):
    ''' Return colormap from string (eg 'jet')'''
    try:
        cmap=getattr(cm, color)
    except AttributeError:
        raise badvalue_error(color, 'a custom LInearSegmentedColormap or a string matching \
        the default colormaps available to matplotlib (http://dept.astro.lsa.umich.edu/~msshin/science/code/matplotlib_cm/)') 
    return cmap


def diag_line(ax, **linekwds):
    """ Draw a diagonal line x=y"""
    linekwds.setdefault('ls', ':')   #Dotted ... 
    linekwds.setdefault('color', 'k') #black
    linekwds.setdefault('linewidth', 1)

    # Bisecting line
    ax.plot(ax.get_xlim(), 
            ax.get_ylim(), 
            **linekwds)    
    return ax

def multi_axes(count, **kwargs):
    """ """
    figsize = kwargs.pop('figsize', None)#, rcParams['figure.figsize'])
    ncols = kwargs.pop('ncols', cnfg.multicols)

    if count <= ncols:
        nrows = 1
        ncols = count

    else:  
#       ncols = _mod_closest(count)
        nrows = int(count/ncols)         
        if count % ncols: #If not perfect division
            nrows += 1

    if figsize:
        fig, axes = splot(nrows, ncols, figsize=figsize, fig=True)
    else:
        fig, axes = splot(nrows, ncols,fig=True)


    while len(fig.axes) > count:
        fig.delaxes(fig.axes[-1])
    return fig, fig.axes, kwargs



def _parse_names(names, default_names):
    """ Boilerplate: user enters *names to overwrite X default names.  
    For example, if user enters 2 names but 5 unique labels in an image 
    are found, and vice versa."""

    default_names = list(default_names)
    names = list(names)

    # Handle various cases of names/values not being the same
    if names:
        if len(names) == len(default_names):
            pass

        elif len(names) < len(default_names):
            logger.warn("length : %s names provided but %s unique "
                        "labels were found" % (len(names), len(default_names)) )              
            default_names[0:len(names)] = names[:]
            return default_names

        else: #len(names) >= len(default_names)
            logger.warn("length : %s names provided but %s unique "
                        "labels were found" % (len(names), len(default_names)) )     
            return names[0:len(default_names)] 

    else:
        names[:] = default_names[:]

    return names

def _annotate_mappable(df, cmap, axis=0, vmin=None, vmax=None):

    if isinstance(cmap, basestring): 
        cmap=cmget(cmap)

    if axis != 0 and axis != 1:
        raise badvalue_error(axis, 'integers 0 or 1')

    # Min and max values of color map must be min and max of dataframe
    if not vmin:
        vmin=min(df.min(axis=axis))
    if not vmax:        
        vmax=max(df.max(axis=axis))

    cNorm = mplcolors.Normalize(vmin=vmin, vmax=vmax)
    scalarmap = cm.ScalarMappable(norm=cNorm, cmap=cmap)    
#   http://stackoverflow.com/questions/6600579/colorbar-for-matplotlib-plot-surface-command
    scalarmap.set_array(np.arange(0,1)) #Thsi can be anything, arbitrary
    return scalarmap, vmin, vmax


def _df_colormapper(df, cmap, axis=0, colorbymax=False, vmin=None, vmax=None):
    ''' Maps matplotlibcolors to a dataframe based on the mean value of each curve along that
    axis.  

    Parameters
    ----------
    colorbymax : bool (False)
        If true, curves are colored based on their maxima value (ie largest
        curve is max color regardless of where it appears in time). Otherwise,
        curves are colored chornologically.

    Notes
    -----
    Useful for df.plot() which doesn't take a normalized colormap natively. 
    cmap can be an instance of an RGB color map, or a string which such that 
    cm.string will produce one.
    '''

    if isinstance(cmap, basestring): 
        cmap=cmget(cmap)

    if axis != 0 and axis != 1:
        raise badvalue_error(axis, 'integers 0 or 1')

    # Min and max values of color map must be min and max of dataframe
    if not vmin:
        try:
            vmin = min(df.min(axis=axis)) 
        except TypeError: #If df only is one column...
            vmin = df.min(axis=axis)
    if not vmax:   
        try:
            vmax = max(df.max(axis=axis))
        except TypeError:
            vmax = df.max(axis=axis)

    cNorm = mplcolors.Normalize(vmin=vmin, vmax=vmax)
    scalarmap = cm.ScalarMappable(norm=cNorm, cmap=cmap)    

    if axis == 0:
        if colorbymax:
            colors=[scalarmap.to_rgba(df[x].max()) for x in df.columns]
        else:
            colors = [scalarmap.to_rgba(x) for x in 
                      np.linspace(vmin, vmax, len(df.columns))]

    elif axis == 1:
        if colorbymax:
            colors=[scalarmap.to_rgba(df.ix[x].max()) for x in df.index]                
        else:
            colors = [scalarmap.to_rgba(x) for x in 
                      np.linspace(vmin, vmax, len(df.index))]

    return colors

def to_normrgb(color):
    """ Returns an rgb len(3) tuple on range 0.0-1.0 with several input styles; 
        wraps matplotlib.color.ColorConvert.  If None, returns config.PCOLOR by
        default."""

    if color is None:
        color = DEFCOLOR 

    # If iterable, assume 3-channel RGB
    if hasattr(color, '__iter__'):
        if len(color) != 3:
            if len(color) == 4:
                color = color[0:3]
                #   logger.warn("4-channel RGBA recieved; ignoring A channel")
            else:
                raise ColorError("Multi-channel color must be 3-channel;"
                                 " recieved %s" % len(color))
        r, g, b = color
        if r <= 1 and g <= 1 and b <= 1:
            return (r, g, b)

        # Any thing like (0, 255, 30) ... uses 255 as upper limit!
        else:
            r, g, b = map(_pix_norm, (r, g, b) )        
            return (r, g, b)


    if isinstance(color, basestring):
        if color == 'random':
            raise NotImplementedError("random color generation not supported")
#            color = rand_color(style='hex')            
        return _rgb_from_string(color)

    # If single channel --> map accross channels EG 22 --> (22, 22, 22)
    if isinstance(color, int):
        color = float(color)

    if isinstance(color, float):
        if color > 1:
            color = _pix_norm(color)
        return (color, color, color)

    if isinstance(color, bool):
        if color:
            return (1.,1.,1.)
        return (0.,0.,0.)

    raise ColorError(ERRORMESSAGE)        

def _uvvis_colors(df, delim=':'):
    '''    From a dataframe with indicies of ranged wavelengths (eg 450.0:400.0), and builds colormap
    with fixed uv_vis limits (for now 400, 700).  Here are some builtin ones:
    http://dept.astro.lsa.umich.edu/~msshin/science/code/matplotlib_cm/.  

    Colors each curve based on the mid value in range. '''
    colors=[]
    cNorm=mplcolors.Normalize(vmin=350.0, vmax=700.0)
    scalarmap=cm.ScalarMappable(norm=cNorm, cmap=cm.jet)  

    for rng in df.index:
        start,stop=rng.split(delim)
        colors.append(scalarmap.to_rgba(0.5 * (float(stop)+float(start) ) ) )
    return colors


def splot(*args, **kwds):
    """ Wrapper to plt.subplots(r, c).  Will return flattened axes and discard
    figure.  'flatten' keyword will not flatten if the plt.subplots() return
    is not itself flat.  If flatten=False and fig=True, standard plt.subplots
    behavior is recovered."""

    flatten = kwds.pop('flatten', True)
    _return_fig = kwds.pop('fig', False)

    fig, args = plt.subplots(*args, **kwds)

    # Seems like sometimes returns flat, sometimes returns list of lists
    # so either way I flatten    
    if not hasattr(args, '__iter__'):
        args = [args]

    try:
        args = [ax.axes for ax in args] 
    except Exception:
        if flatten:
            args = [ax.axes for row in args for ax in row]
        else:
            args = [tuple(ax.axes for ax in row) for row in args]

    if _return_fig:
        return (fig, args)
    else:
        return args


def hide_axis(ax, axis='x', axislabel=True, ticklabels=True, ticks=False,
              hide_everything=False):
    """ Hide axis features on an axes instance, including axis label, tick
    labels, tick marks themselves and everything.  Careful: hiding the ticks
    will also hide grid lines if a grid is on!

    Parameters
    ----------

    axis : 'x' or 'y' 'both'

    axislabel : True
        Hide the axis label

    ticklabels : True
        Hide the tick labels

    ticks : True
        Hide the tick markers

    hide_everything : False
        Hides labels, tick labels and axis labels.

    """
    if axis not in ['x','y','both']:
        raise AttributeError('axis must be "x" or "y" or both')

    axis_to_modify = []

    if hide_everything:
        ticks = True; axislabel=True; ticklabels=True

    # Set axis labels
    if axis == 'x' or axis == 'both':
        axis_to_modify.append(ax.get_xaxis())

        if axislabel: 
            ax.set_xlabel('')

    if axis == 'y' or axis == 'both': #not elif becuase "both" stipulation
        axis_to_modify.append(ax.get_yaxis())

        if axislabel:
            ax.set_ylabel('')

    for an_axis in axis_to_modify:
        if ticklabels:
            an_axis.set_ticklabels([])
        if ticks:
            an_axis.set_ticks([])

    return ax 


def invert_ax(ax):
    """ Inverts x and y data on a plot axis, and flips limits.  Most useful for
    having an oriented axes in correlation side plots.
    """
    for line in ax.lines:
        xd = line.get_xdata()
        yd = line.get_ydata()
        line.set_xdata(yd)
        line.set_ydata(xd)

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    ylim = (ylim[-1], ylim[0])
    ax.set_xlim(ylim)
    ax.set_ylim(xlim)
    return ax


def easy_legend(ax, fancy=True, position='top', **legendkwds):
    ''' Wrapper around ax.legend to make it easy to move a legend around the edges of the plot.  Made for sensible
    numbers of lines (1-15) and tries to choose smart placement to avoid conflict with lines, labels etc...

    BROKEN!!!

    This is a bust, since plotting legend is easy enough.  See guide, especially with the 'loc'      
    http://matplotlib.org/users/legend_guide.html

    If coming back to this, here are the issues to resolve:
       legend['loc'] must be enabled for the bounding box to work correctly/consistently.
       bbox coordinates are (left edge, bottom edge) of legend.  
         -controlling bottom edge is really dumb because want the top edge to clear the xlabel for example, and bottom
           edge depends on thickness of plot.  Putting the legend on top actually works nicely for this.

       If plotting outside left/right of plot, need to squeeze the plot in afterall, it will not accommadate my legend.
       Had this code, but deleted it.  Basically, need to squeeze left/right width of the axes.bounding width by 20%
       per column of legend.  
       (see second reply here http://stackoverflow.com/questions/4700614/how-to-put-the-legend-out-of-the-plot)


    '''

    ir=lambda x: int(round(x))
    position=position.lower() 
    legendkwds['loc']=3 #MUST BE SET TO SOMETHING, 2 or 3 doesn't matter, or lower left is no longer start point    

    if position not in ['top', 'bottom', 'left', 'right']:
        raise badvalue_error(position, 'top, bottom, left, or right')

    ################################################################
    ### Choose skinny legend for top bottom, long for left/right ###
    ################################################################
    if 'ncol' not in legendkwds:
        if position=='top' or position=='bottom':

            if len(ax.lines) < 4:
                ncol=len(ax.lines)
            else:
                ncol=4           
        else:
            ncol=ir( len(ax.lines)/20.0) #For left right, nice if columns have large vertical extent (10 lines is good)
            if ncol==0:
                ncol=1
        legendkwds['ncol']=ncol

    ###################################################
    ### Choose legend position around plot elements ###
    ###################################################

    box=ax.get_position()   #Can do box.x0 and stuff when resetting ax.set_position()

    if 'bbox_to_anchor' not in legendkwds:  #Don't use has_key(), gone in 3.x

        if position=='top':     
            ''' Legend slightly inside the upper bound of plot'''
            if ax.get_title() == '':
                bbox_to_anchor=(0.2, 1.025)  #0.25 will only center plot if plot is 0.5 units long, then
            else:                             #other end at 0.75.  0.2 looks nice for 8 column uv-vis plot. 
                bbox_to_anchor=(0.2, 1.05)   #Also depends on size of monitor!! so either way, screwed


        elif position=='bottom':
            ''' Centered legend under the label'''

            if ax.get_xlabel()=='':
                bbox_to_anchor=(0.25, -0.025)
            else:
                bbox_to_anchor=(0.25, -0.05)


        elif position=='right':
            ''' Squeeze 20% width inward per column in legend to make room'''
            bbox_to_anchor=(1.05, 0.0) #WHY DOESNT 1,1 work

        elif position=='left':
            ''' Squeeze 20% width inward per column in legend to make room'''
            if ax.get_ylabel()=='':
                bbox_to_anchor=(-0.07, 1.0)   
            else:
                bbox_to_anchor=(-0.12, 1.0)


        legendkwds['bbox_to_anchor']=bbox_to_anchor

    if fancy and 'fancybox' not in legendkwds and 'shadow' not in legendkwds:
        legendkwds['fancybox']=True
        legendkwds['shadow']=True

    if 'borderaxespad' not in legendkwds:
        legendkwds['borderaxespad']=0.0  #Havne't played with this

        ### WHY IS MODE BROKEN (comes up garbled on plot)  
#    if 'mode' not in legendkwds:
    #       legendkwds['mode']='expand'


    ax.legend(**legendkwds)
    return ax


"""
Adapted to fit skspec 5/6/2013.  Original credit to Alexis Mignon:

Module for Principal Component Analysis.

Features:
* pca and kernel pca
* pca through singular value decomposition (SVD)

Author: Alexis Mignon (c)
Date: 10/01/2012
e-mail: alexis.mignon@gmail.com
"""

import numpy as np
from scipy.sparse.linalg.eigen.arpack import eigs
from scipy.linalg import eigh

def full_pca(data):
    """
    Performs the complete eigen decomposition of
    the covariance matrix.
    
    arguments:
    * data: 2D numpy array where each row is a sample and
        each column a feature.
    
    return:
    * w: the eigen values of the covariance matrix sorted in from 
          highest to lowest.
    * u: the corresponding eigen vectors. u[:,i] is the vector
         corresponding to w[i]
         
    Notes: If you want to compute only a few number of principal
           components, you should consider using 'pca' or 'svd_pca'.
    """
    cov = np.cov(data.T)
    w,u = eigh(cov,overwrite_a = True)
    return w[::-1],u[:,::-1]

def pca(data,k):
    """
    Performs the eigen decomposition of the covariance matrix.
    
    arguments:
    * data: 2D numpy array where each row is a sample and
            each column a feature.
    * k: number of principal components to keep.
    
    return:
    * w: the eigen values of the covariance matrix sorted in from 
          highest to lowest.
    * u: the corresponding eigen vectors. u[:,i] is the vector
         corresponding to w[i]
         
    Notes: If the number of samples is much smaller than the number
           of features, you should consider the use of 'svd_pca'.
    """
    cov = np.cov(data.T)
    # kw "which" means return largest magnitude k eigenvalues
    w,u = eigs(cov,k = k,which = 'LM')
#    return w[::-1],u[:,::-1]
    return w,u #(No need to reverse w,u because eigs does it using 'LM')

def extern_pca(data,k):
    """
    Performs the eigen decomposition of the covariance matrix based
    on the eigen decomposition of the exterior product matrix.
    
    
    arguments:
    * data: 2D numpy array where each row is a sample and
            each column a feature.
    * k: number of principal components to keep.
    
    return:
    * w: the eigen values of the covariance matrix sorted in from 
          highest to lowest.
    * u: the corresponding eigen vectors. u[:,i] is the vector
         corresponding to w[i]
         
    Notes: This function computes PCA, based on the exterior product
           matrix (C = X*X.T/(n-1)) instead of the covariance matrix
           (C = X.T*X) and uses relations based of the singular
           value decomposition to compute the corresponding the
           final eigen vectors. While this can be much faster when 
           the number of samples is much smaller than the number
           of features, it can lead to loss of precisions.
           
           The (centered) data matrix X can be decomposed as:
                X.T = U * S * v.T
           On computes the eigen decomposition of :
                X * X.T = v*S^2*v.T
           and the eigen vectors of the covariance matrix are
           computed as :
                U = X.T * v * S^(-1)
    """
   
#    raise NotImplementedError('Need to curate this method')
    # Should I take last line and remove the [::-1] stuff?
    data_m = data - data.mean(0)
    K = np.dot(data_m,data_m.T)
    w,v = eigs(K,k = k,which = 'LM')
    U = np.dot(data.T,v/np.sqrt(w))
    # Normalizes eigenvalues by length of data (?)
#    return w[::-1]/(len(data)-1),U[:,::-1]
    return w/(len(data)-1),U

def full_kpca(data):
    """
        Performs the complete eigen decomposition of a kernel matrix.
        
        arguments:
        * data: 2D numpy array representing the symmetric kernel matrix.
        
        return:
        * w: the eigen values of the covariance matrix sorted in from 
              highest to lowest.
        * u: the corresponding eigen vectors. u[:,i] is the vector
             corresponding to w[i]
             
        Notes: If you want to compute only a few number of principal
               components, you should consider using 'kpca'.
    """
    w,u = eigh(data,overwrite_a = True)
#    return w[::-1],u[:,::-1]
    return w,u


def kpca(data,k):
    """
        Performs the eigen decomposition of the kernel matrix.
        
        arguments:
        * data: 2D numpy array representing the symmetric kernel matrix.
        * k: number of principal components to keep.
        
        return:
        * w: the eigen values of the covariance matrix sorted in from 
              highest to lowest.
        * u: the corresponding eigen vectors. u[:,i] is the vector
             corresponding to w[i]
             
        Notes: If you want to perform the full decomposition, consider 
               using 'full_kpca' instead.
    """
    w,u = eigs(data,k = k,which = 'LM')
#    return w[::-1],u[:,::-1]
    return w,u

class PCA(object):
    """
        PCA object to perform Principal Component Analysis.
    """
    def __init__(self,k = None, kernel = False, extern = False):
        """
        Constructor.
        
        arguments:
        * k: number of principal components to compute. 'None'
             (default) means that all components are computed.
        * kernel: perform PCA on kernel matrices (default is False)
        * extern: use extern product to perform PCA (default is 
               False). Use this option when the number of samples
               is much smaller than the number of features.

        Notes:
        * All data will be mean-cenetered.  Np subroutines (eg np.cov())
          do this in all cases except for the extern_pca() method, which
          does this automatically.
        """
    
        self._k = k #THESE SHOULD BE PROPERTIES
        self._kernel = kernel
        self._extern = extern
        
    
    def fit(self,X):
        """
        Performs PCA on the data array X.
        arguments:
        * X: 2D numpy array. In case the array represents a kernel
             matrix, X should be symmetric. Otherwise each row
             represents a sample and each column represents a
             feature.
        """
        
        # Mean centering is inherently performed in numpy methods that compute
        # the covariance.  This does it additionally because X may not be
        # the same matrix used to fit the original data
        
        # If number principle componenets is not specified, full pca or kpca
        if self._k is None :
            if self._kernel :
                pca_func = full_kpca
            else :
                pca_func = full_pca
            self.eigen_values_,self.eigen_vectors_ = pca_func(X)

        # If specifying number of principle components
        else :
            if self._kernel :
                pca_func = kpca
            elif self._extern :
                pca_func = extern_pca
            else :
                pca_func = pca
            self.eigen_values_,self.eigen_vectors_ = pca_func(X,self._k)
        
        if self._kernel :
            
            total_variance = X.diagonal().sum()

            
        # If not kernal pca, subtract mean.
        # Total variance is squared difference from mean, since mean centered.
        # Simply computes variance of each element
        else :
            self.mean = X.mean(0)
            diff = X - self.mean
            total_variance = (diff*diff).sum()/(X.shape[0]-1)

        # Scikit image also has this; how much variance each component
        # can account for.  Should sum to 1 if all components used.
        self.explained_variance_ = self.eigen_values_.sum()/total_variance
        return self
    
        
    def transform(self,X,whiten = False):
        """
        Project data on the principal components. If the whitening
        option is used, components will be normalized to that they
        have the same contribution.
        
        arguments:
        * X: 2D numpy array of data to project.
        * whiten: (default is False) all components are normalized
            so that they have the same contribution.
            
        returns:
        * prX : projection of X on the principal components.
        
        Notes: In the case of Kernel PCA, X[i] represents the value
           of the kernel between sample i and the j-th sample used
           at train time. Thus, if fit was called with a NxN kernel
           matrix, X should be a MxN matrix.
           
           The projection in the kernel case is made to be equivalent
           to the projection in the linear case.
           
               X.T = U * S * v.T
               C = 1/(N-1) * X.T * X
               X.T * X = U*S^2*U.T
               K = X * X.T = v*S^2*v.T
               
               U = X.T * v * S^(-1)
           
           The projection with PCA is :
               X' = X * U
               X' = X * X.T * v * S^(-1)
               X' = K * v * S^(-1)
               
           For whiten PCA :
               X' = X * U * S^(-1) * sqrt(N-1)
               X' = X * X.T * v * S^(-1) * S^(-1) * sqrt(N-1)
               X' = K * S^(-2) * sqrt(N-1)
        """
    
        
        if self._kernel :
            pr = np.dot(X,self.eigen_vectors_)
            if whiten :
                pr /= self.eigen_values_ / np.sqrt(X.shape[0]-1)
            else :
                pr /= np.sqrt(self.eigen_values_)
        else :
            pr = np.dot(X - self.mean,self.eigen_vectors_)
            if whiten:
                pr /= np.sqrt(self.eigen_values_)
        return pr
    
        


"""Provides core "TimeSpectra" class and associated utilities."""

import logging
from skspec.logger import decode_lvl, logclass
from pandas import DatetimeIndex, Index
from skspec.core.spectra import _valid_xunit
from skspec.units.intvlunit import INTVLUNITS, DatetimeCanonicalError
from skspec.core.timeindex import TimeIndex
from spectra import Spectra

logger = logging.getLogger(__name__) 

def _valid_intvlunit(sout):
    """ Validate interval unit."""
    return _valid_xunit(sout, intvl_dic)

# Ignore all class methods!
#@logclass(log_name=__name__, skip = ['wraps','_framegetattr', 'from_csv', 
 #                                    '_comment', '_transfer'])
class TimeSpectra(Spectra):
    """ Spectra that enforces a TimeIndex.  TimeIndex converts representations
    between DatetimeIndex and Intervals.

    Parameters
    ----------

    force_datetime: bool (False)
        Convienence method for when a user is passing in datetimes, but does
        not have them in a DatetimeIndex.  The only use case we've found is
        trying to parse CSV files; the datetimes are a list of strings and
    """


    # Has to be this way because class methods not accessible via metadataframe __getattr__()
    def __init__(self, *dfargs, **dfkwargs):
        dfkwargs.setdefault('strict_columns', TimeIndex)            
        super(TimeSpectra, self).__init__(*dfargs, **dfkwargs)
        


## TESTING ###
if __name__ == '__main__':
    from specindex import SpecIndex
    from pandas import date_range
    import numpy as np
    from skspec.data import solvent_evap, aunps_glass, aunps_water
    
    x = AnyFrame(np.random.randn(50,50))
    print x

    def sumdiff(array, absolute=False, cumsum=False):
        """ Sum of the differences of an array (usually array of areas).  If absolute, absolute difference is used.
        If cumsum, the cumulative sum is returned.
        """
        sdiff = []
        # Compute the forward difference up to i=final -1
    #    array = array.values
        for idx, value in enumerate(array):
            if idx < len(array)-1:
                new = array[idx+1]
                out = new-value          
                if absolute:
                    out = abs(out)        
                sdiff.append(out)
    
        if cumsum:
            sdiff = list(np.cumsum(sdiff)) 
        sdiff.append(sdiff[-1]) #REPEAST LAST VALUE
        return sdiff

    ts = aunps_glass()

    ts.iunit = None #is the most accurate
    ts.varunit = 'm'
    from pandas import Index
    ts.columns = Index(ts.columns)
    area = ts.area()
    
    DIFF = area.apply(sumdiff, axis=1, raw=False)
    

    # Be careful when generating test data from Pandas Index/DataFrame objects, as this module has overwritten their defaul behavior
    # best to generate them in other modules and import them to simulate realisitc usec ase


    ## For testing 
    #import matplotlib.pyplot as plt


    #from skspec.data import aunps_water, aunps_glass
    #from skspec.plotting import splot, range_timeplot
    #ts_water = aunps_glass()
##    ax1, ax2 = splot(1,2)

##    ts_water.plot(ax=ax1)
    #range_timeplot(ts_water.wavelength_slices(8))
##    bline = ts_water.baseline
##    bline.plot(ax=ax1, lw=5, color='r')
    #plt.show()



    spec=SpecIndex(range(400, 700,1), unit='nm')
###    spec=SpecIndex([400.,500.,600.])
    testdates = date_range(start='3/3/12',periods=30,freq='h')
    ##testdates2 = date_range(start='3/3/12',periods=30,freq='45s')

    #ts=TimeSpectra(abs(np.random.randn(300,30)), 
##                   columns=testdates, 
                    #index=spec, 
                    #name='ts1')  

    #print ts.list_varunits()
    #print ts.list_specunits()

    ###t2=TimeSpectra(abs(np.random.randn(300,30)), 
                    ###columns=testdates2, 
                    ###index=spec, 
                    ###name='ts2') 

    import matplotlib.pyplot as plt
    from skspec.plotting import areaplot
    ts = solvent_evap()
    ts = ts.as_varunit('m')

    from pandas import DataFrame
    df = DataFrame(np.random.randn(50,50), index=np.linspace(0,50,50))
    
#    ts.nearby[900:]
    ts.nearby[[1200,1400]]
        
    ts.iloc[:, 2:4.4].columns
    ts._frame.iloc[:, 2:4.4].columns
#    ts = aunps_glass()
    print ts.ix[1505.0:1500.0]
    #ts.loc[:, 50.5, 520.0]
    ##area = ts.area()
    ##tf=ts[ts.columns[-1]]
    ##print tf 
    #cols = ts.columns[0:5]
    #c2 = ts.columns[0:5].convert('m')
    #c2.datetimeindex
    #mins = cols.convert('m')
    #print mins
    #ts.as_varunit('m')

    #t2 = ts.ix[1500.0:1000.0]
    #print ts.index
    #print t2.index


    ##t2 = ts.as_interval('m')

    ##t2 = t2.as_iunit('r')

    ###stack = ts.split_by(1)
    ###stack.iunit = 'a'

    ###ts[ts.columns[0]].plot(colormap='RdBu')
    ###plt.show()
    ##t2 = ts.as_interval('m')
    ###t2 = ts.as_iunit('r')
    ###ts.area().plot()
    ###import sys
    ###sys.exit()

####    stack.plot(title='Big bad plots')
    ##from skspec.plotting import six_plot
    ##import matplotlib.pyplot as plt
    ##six_plot(ts, striplegend=True)
    ##plt.show()
    ###t1 = ts.as_interval()
    ###print t1.columns
    ###t1.plot(cbar=True)
    ###t1.to_datetime()
    ###t1.ix[500.0:600.0]
    ###t2 = ts.as_specunit('ev')
    ###t3 = ts.as_iunit('a')
    ###print t2.specunit, 'hi t2'
    ###print t3.specunit, 'hi t3'
    ###print t2.specunit, 'hi t2'
    ###specplot(ts, cbar=True)



####    a=ts.area()
####    print 'hi', a.specunit
####    ts.specunit = 'ev'
    ####from skspec.plotting import specplot, areaplot
    ####areaplot(ts)
    ####plt.show()

    ####from skspec.IO.gwu_interfaces import from_spec_files, get_files_in_dir
    ####from skspec.exampledata import get_exampledata
    ####ts=from_spec_files(get_files_in_dir(get_exampledata('NPSAM'), sort=True), name='foofromfile')

    ###ts.to_interval('s')
    ###ts=ts.ix[440.0:700.0,0.0:100.0]
    ###ts.reference=0    
    ###print ts._baseline.shape, ts.shape

    #### Goes to site packages because using from_spec_files, which is site package module
    ###ts.run_pca()
    ####   ts.pca_evals

    ####from pandas import Panel
    ####Panel._constructor_sliced=TimeSpectra
    ####pdic={'ts':ts}
    ####tp=Panel.from_dict(pdic)

    ###d={'start':2/22/12, 'periods':len(ts.columns), 'freq':'45s'}
    ###ts.set_daterange(start='2/22/12', periods=len(ts.columns), freq='45s')

    ###ts.baseline=ts.reference
    ###ts.sub_base()

    #### THIS FAILS WHEN INDEX=SPEC 
    ###t3=TimeSpectra(abs(np.random.randn(ts.baseline.shape[0], 30)), columns=\
                    ###testdates, 
                    ###baseline=ts._baseline, name='foobar')  



    ####ts._reference.x='I WORK'
    ####ts._reference.name='joe'
    ###ts.baseline=Series([20,30,50,50], index=[400., 500., 600., 700.])
#####    t2.baseline=ts.baseline
    ####ts._frame.ix[:, 0:4]
    ####ts.ix[:,0:4]
    ####ts.pvutils.boxcar(binwidth=20, axis=1)
    ####x=ts.ix[450.0:650.]
    ####y=t2.ix[500.:650.]

    ####ts.cnsvdmeth='name'

    ###from skspec.pandas_utils.metadframe import mload
    ####from skspec import areaplot, absplot
    ###ts=mload('rundata.pickle')    
    ###ts=ts.as_interval('m')
    ###x=ts.area()    
    ###print 'hi', x
    ####ts.reference=0
    ####ts=ts[ts.columns[800.0::]]
    ####ts=ts.ix[400.0:800.0]
    ####c=haiss_m2(ts, peak_width=2.0, ref_width=2.0)
    ####a=haiss_m3(ts, 0.000909, peak_width=None, dilution=0.1)
    ####b=haiss_conc(ts, 12.0)
    #####b2=haiss_conc(ts, 12.0, dilution=0.2)

#####    bline=ts[ts.columns[0]]
#####    ts=ts.ix[:,25.0:30.0]
#####    ts.reference=bline

    ####uv_ranges=((430.0,450.0))#, (450.0,515.0), (515.0, 570.0), (570.0,620.0), (620.0,680.0))

    ####tssliced=ts.wavelength_slices(uv_ranges, apply_fcn='mean')

    ####from skspec.core.utilities import find_nearest
    ####x=ts.ix[500.:510, 0]
    ####b=pvutils.maxmin_xy(x)
    ####a=find_nearest(x, .15)
    ####ts.iunit=None
    ####ts.iunit='a'
    ####ts.iunit=None


    ####ts.to_csv('junk')
    ####range_timeplot(ts)

    ####ts.a=50; ts.b='as'
    ####ts.iunit='t'

    ####ts.list_attr(dfattr=False, methods=False, types=True)    
    ####ts.as_iunit('a')
    ####x=ts.as_iunit('a')
    #####ts.as_interval()
    #####spec_surface3d(ts)  ##Not working because of axis format problem
    #####plt.show()
#####    ts.rank(use_base=True)
    ####x=ts.dumps()
    ####ts=mloads(x)


""" GWU in-house script for dataanalysis of fiberoptic probe data."""

__author__ = "Adam Hughes"
__copyright__ = "Copyright 2013, GWU Physics"
__license__ = "Free BSD"
__maintainer__ = "Adam Hughes"
__email__ = "hugadams@gwmail.gwu.edu"
__status__ = "Development"

import os
import os.path as op
import datetime as dt 
import shutil
import sys
import imp
import shlex
import logging
import argparse
import matplotlib.pyplot as plt
import numpy as np
from collections import OrderedDict
from time import gmtime, strftime

# skspec IMPORTS
#from skspec.bundled import run_nb_offline
from skspec.plotting import areaplot, range_timeplot, six_plot
#from skspec.plotting.advanced_plots import spec_surface3d, surf3d, spec_poly3d, plot2d, plot3d
from skspec.core.utilities import boxcar, countNaN
from skspec.core.baseline import dynamic_baseline
from skspec.plotting.plot_utils import _df_colormapper, cmget
from skspec.IO.gwu_interfaces import from_timefile_datafile, from_spec_files
from skspec.core.file_utils import get_files_in_dir, get_shortname
#from skspec.core.corr import ca2d, make_ref, sync_3d, async_3d
from skspec.core.spectra import _normdic
from skspec.pandas_utils.metadframe import mload, mloads
from skspec.exceptions import badkey_check, ParserError, GeneralError, \
     ParameterError, LogExit

from skspec.scripts.gwu_script.tex_templates import SECTION

# Import some spectrometer configurations
from parameters_model import Parameters, USB2000, USB650

logger = logging.getLogger(__name__)
from skspec.logger import configure_logger, logclass
from skspec.data import data_dir

SCRIPTNAME = 'gwuspec'
DEF_INROOT = './scriptinput'
DEF_OUTROOT = './output'
ALL_ANAL = ['1d', '2d', '3d', 'corr']   
ANAL_DEFAULT = ['1d']#, '2d']

# HOW TO READ THIS ABSOLUTE PATH
IPYNB = op.join(data_dir, '_script_nb.ipynb')  #IPYTHON NOTEBOOK TEMPLATE
NBVIEWPATHS = [] #Where various ipython notebook blob links are stored

def hideuser(abspath):
    """ Replace user home directory with ~.  Inverse operation for
    op.expanduser().  Must pass absolute path that starts with user.
    User is home/glue but on FD harddrive, it's media/backup, so hack
    to just replace
    """
    user = op.expanduser('~')
    if '/media/backup' in abspath:
        user = '/media/backup'
    return abspath.replace(user, '~')
    

def ext(afile): 
    """ get file extension"""
    return op.splitext(afile)[1]

def timenow():
    return dt.datetime.now()

def logmkdir(fullpath):
    """ Makes directory path/folder, and logs."""
    logger.info('Making directory: %s' % fullpath)
    os.mkdir(fullpath)
    
def latex_string(string):
    """ Replace underscore, newline, tabs to fit latex."""

    string = string.replace('_', '\\_')
#    string = string.decode('string-escape')
    
    string = string.replace('\n', '\\\\')
    return string.replace('\t', '\hspace{1cm}') #Chosen arbitrarily

def dict_out(header, dic, sort = True):
    """ Retures a string of form header \n k \t val """
    
    if sort:
        return header + ':\n\n' + '\n'.join(['\t' + (str(k) + '\t' + str(v)) 
                 for k,v in sorted(dic.items())])
    else:
        return header + ':\n\n' +'\n'.join(['\t' + (str(k) + '\t' + str(v)) 
                 for k,v in dic.items()])        
    
def latex_multicols(dic, title='', cols=2):
    """ Hacky: given a list of items, wraps a multicol iterable so that the list
        will be output as columns.  Title will be added in bold/large"""
    
    outstring = ''
    if title:
        outstring += r'{\bf \large %s}' % title

    outstring += r'\begin{multicols}{%s}\begin{itemize}' % cols
    for k, v in dic.items():
        outstring += r'\item{\ttfamily {\bf %s}:\hspace{.3cm}%s}' % (k, v)
    outstring += '\end{itemize}\end{multicols}'
    return latex_string(outstring)
    
    

# This could delete empty folders that were otherwise in the directory that 
# were there before starting script
def removeEmptyFolders(path):
    """ Removes any empty folders in path.  Useful because recursive mode
        of script makes every directory it finds, even if they are empty."""
    if os.path.isdir(path):
  
        # remove empty subfolders
        if os.listdir(path):
            for f in os.listdir(path):
                fullpath = os.path.join(path, f)
                if os.path.isdir(fullpath):
                    removeEmptyFolders(fullpath)
      
        else: 
            logger.info('Removing empty folder: "%s"' % path)
            os.rmdir(path)


class AddUnderscore(argparse.Action):
    """ Adds an underscore to end of value (used in "rootname") """
    def __call__(self, parser, namespace, values, option_string=None):
        if values != '':
            values += '_'
        setattr(namespace, self.dest, values)
    

#skip class methods
@logclass(log_name=__name__ , public_lvl='debug',
          skip=['_ts_from_picklefiles', 'from_namespace'])
class Controller(object):
    """ """

    name = 'Controller' #For logging

    # Extensions to ignore when looking at files in a directory       
    img_ignore=['.png', '.jpeg', '.tif', '.bmp', '.ipynb'] 
        
    nb_links = {} #Stores ipython notebook viewer links    
    
    # For now, this only takes in a namespace, but could make more flexible in future
    def __init__(self, **kwargs):
              
        # These should go through the setters        
        self.inroot = kwargs.get('inroot', DEF_INROOT) 
        self.outroot = kwargs.get('outroot', DEF_OUTROOT)
        self.sweepmode = kwargs.get('sweep', False)
        self.analysis = kwargs.get('analysis', ANAL_DEFAULT)

        self._plot_dpi = kwargs.get('plot_dpi', None) #Defaults based on matplotlibrc file
        self._plot_dim = kwargs.get('plot_dim', 'width=6cm')
        self._plot_fontsize = kwargs.get('fontsize', 18)
        # Ticksize defaults to 15 btw
    
        
        self.rname = kwargs.get('rname', '')
        self.overwrite = kwargs.get('overwrite', False)
        
        # Configure logger
        verbosity = kwargs.get('verbosity', 'warning')
        trace = kwargs.get('trace', False)
                    
        # Pop kwargs['params'] for output below
        self.params = kwargs.pop('params', None)

        # Add logging later; consider outdirectory conflicts
        self.build_outroot()
        
        configure_logger(screen_level=verbosity, name = __name__,
                 logfile=op.join(self.outroot, 'runlog.txt'), mode='w')
        
        # Output the parameters
        self._run_params_file = op.join(self.outroot, 'runparameters.tex')
        with open(self._run_params_file, 'w') as f:
            
            # Hacky way to write latex section from raw string literals
            f.write(r'\subsection{Script Parameters}')
            f.write(r'\fboxsep5mm')
            f.write(r'\fcolorbox{black}{yellow}{\vbox{\hsize=10cm \noindent \scriptsize ')
  
            f.write(latex_multicols(self.params, title='skspec Parameters'))
            f.write(latex_multicols(kwargs, 'Analysis Parameters'))
            f.write('}}')


        if self._plot_dpi > 600:
            logger.warn('Plotting dpi is set to %s.  > 600 may result in slow'
                       ' performance.')
            
        # Used for storing variables in reports/matlab other uses
        self._run_summary = '' 
        self._csv_paths = []
                        
    def _valid_inpath(self, value):
        if value is None:
            return 
 
        path = op.abspath(value)
        if not op.exists(path):
            raise IOError('Inroot path does not exist: %s' % path)
        return path

    @property
    def inroot(self):
        if not self._inroot:
            raise AttributeError('Root indirectory not set')
        return self._inroot #Abspath applied in setter
    
    @property
    def analysis(self):
        return self._analysis

    @analysis.setter
    def analysis(self, values):
        """ Values must be an iterable. """
        if not values:
            self._analysis = ANAL_DEFAULT
            return
        
        values = list(set(values)) #Remove duplicates
        values = [str(v).lower() for v in values] #String convert 
                
        if 'all' in values or values == ['all']:
            self._analysis = ALL_ANAL
                
        else:
            invalid = [v for v in values if v not in ALL_ANAL]
            if invalid:
                raise ParameterError('Invalid analysis parameter(s) received '
                    '%s.  Choices are %s' % (invalid, ALL_ANAL))
            self._analysis = values
            
                
    @inroot.setter
    def inroot(self, value):
        """ Ensure inpath exists before setting"""
        self._inroot = self._valid_inpath(value)

    @property
    def inpath(self):
        if not self._inpath:
            raise AttributeError('Inpath not set')
        return self._inpath #Abspath applied in setter

    
    @inpath.setter
    def inpath(self, value):
        """ Ensure inpath exists before setting"""
        self._inpath = self._valid_inpath(value)        
        
        
        
    @property
    def outroot(self):
        if not self._outroot:
            raise AttributeError('Root outdirectory not set')
        return self._outroot
        

    @outroot.setter
    def outroot(self, value):
        if value is None:
            self._outroot = None
        else:
            self._outroot = op.abspath(value)
            
 
    @property
    def outpath(self):
        if not self._outpath:
            raise AttributeError('Outpath not set')
        return self._outpath
        

    @outpath.setter
    def outpath(self, value):
        if value is None:
            self._outpath = None
        else:
            self._outpath = op.abspath(value)           
            
        
    @property
    def infolder(self):
        return op.basename(self.inpath)


    @property
    def outfolder(self):
        if not self.outpath:
            raise AttributeError('outfolder not found - outpath not set!')
        return op.basename(self.outpath)


    @property
    def params(self):
        if not self._params:
            raise AttributeError('Spec parameters not set')
        return self._params
    
    
    @params.setter
    def params(self, params):
        """ Can be None, a dict, or an instance of Parameters. """
        if params is None:
            self._params = params
            
        else:
            if isinstance(params, Parameters):
                self._params = params
    
            elif isinstance(params, dict):
                self._params = Parameters(**params)
            else:            
                raise ParameterError('Spectral parameters must be a dictionary '
                      'or parameters_model.Parameters object.  Got object of type %s'
                       % type(params))
            
        
            
    def start_run(self):
        """ Analyzes the root directory.  If sweep mode, subdirectories
            will be analyzed as well."""
        
        self._inpath = self.inroot
        self._outpath = op.join(self.outroot, self.infolder)
        self._treedic = OrderedDict()
        
        self.analyze_dir()
        
        # Remove empty folders that may have been generated during sweep
        if self.sweepmode:
            self.main_walk()   
            removeEmptyFolders(self.outroot)

        # Add main run parameters as section to end of treefile
        self._treedic['Analysis Parameters'] = self._run_params_file

        with open(op.join(self.outroot, 'tree'), 'w') as treefile:
            treefile.write(str(self._treedic))
            
        self.make_matlab(op.join(self.outroot, 'readfiles.m'))
            
        # Add notebooks links to params file    
        logger.info('Appending ipython notebook links to runparams.tex')
        with open(self._run_params_file, 'a') as f:
            # Hacky way to write latex section from raw string literals
            f.write(r'\subsection{IPython Notebooks}')
            f.write(r'\bad{Links will not work until git pushed.}')
            f.write(r'\begin{itemize}')
            for (name, path) in NBVIEWPATHS:
                f.write('\link{%s}{%s}' % (path, latex_string(name)))
            f.write(r'\end{itemize}')     




    def save_csv(self, ts, csv_path, meta_separate=None):
        """ Save timespectra to self.infolder (boilerplate reduction).  Saves
        csv, pickle and metadata.  This is called at least twice, once for 
        full unadalterated data, once for baseline/cropped data. """

        logger.info('Outputting csv to %s.  Metadata will be exluded.' % csv_path)
        ts.to_csv(csv_path, meta_separate=None)
        self._csv_paths.append(csv_path)
        
        
        
    def make_matlab(self, outpath):
        """ Fills simple.m template from mlab_templates with outpath, using
            self._csv_paths."""
        if op.splitext(outpath)[-1] != '.m':
            raise IOError('matlab file must end in ".m". '  
            'Received "%s"' % outpath)

        logger.info('Making matlab script: "%s"' % outpath)
        from skspec.scripts.gwu_script.mlab_templates import SIMPLE_M

        m_dic={}
        # Do some hacky string formatting to fit matlab code
        m_dic['files'] = ',\n'.join("'"+item+"'" for item in self._csv_paths)
        basenames = [op.splitext(op.basename(path))[0] for path in self._csv_paths]
        m_dic['basenames'] = ',\n'.join("'"+item+"'" for item in basenames)
        m_dic['attrnames'] = '\n'.join(['%s=myData.%s' % (i,i) for i in basenames])

        
        mfile = open(outpath, 'w')
        mfile.write( SIMPLE_M % m_dic )                        
        mfile.close()        
        

    def main_walk(self):
        """ Walks all the subdirectories of self.inroot; runs analyze_dir()"""
        
        walker = os.walk(self.inroot, topdown=True, onerror=None, followlinks=False)
        (rootpath, rootdirs, rootfiles) = walker.next()   

        rootdirs.sort() #Iterate alphebetically
        
        if not rootdirs:
            logger.warn('Recursive walk found no further directories after %s'
                        % self.infolder)    
    
        while rootdirs:
            
            # if npsam is in folder name, put that folder first.  
            for idx, folder in enumerate(rootdirs):
                if 'npsam' in folder.lower():
                    rootdirs.pop(idx)
                    rootdirs.insert(0, folder)
                    logger.info('Found "npsam" matching directory in folder "%s"'
                                ' resorting alphebatized directories.')
                    break                

            for iteration, folder in enumerate(rootdirs):
                                
         # Outsuffix is working folder minus inroot (inroot/foo/bar --> foo/bar)
                wd = op.join(rootpath, folder)
                outsuffix = wd.split(self.inroot)[-1].lstrip('/') 
                
                self._inpath = wd
                self._outpath = op.join(self.outroot, outsuffix)
                
                self.analyze_dir()

        # Move down to next directory.  Need to keep iterating walker until a
        # directory is found or until StopIteraion is raised.
            rootpath, rootdirs, rootfile=walker.next()
            if not rootdirs:
                try:
                    while not rootdirs:
                        rootpath, rootdirs, rootfile=walker.next()
                except StopIteration:
                    logger.info('Reached end of directory tree.')
                    break


    def analyze_dir(self):
        """ Wraps loging to self._analyze_dir """
        
        logger.debug('inpath is: %s' % self._inpath)
        logger.debug('outpath is: %s' % self._outpath)        
        try:
            self._analyze_dir()
        except LogExit: #log exit
            logger.critical('FAILURE: "%s" finished with errors.' % self.infolder) 
        else:
            # Generate report
            logger.info('SUCCESS: "%s" data analysis complete' % self.infolder)       
            self.section_report()
                        

    def section_report(self):
        """ Writes a section report for the current inpath.
            These are tracked via the treefile, for compatability
            with specreport.py"""
        
        # path/to/section/ ---> section [and latex formatted for "_" 
        secname = latex_string(self.outpath.split(self.outroot)[-1].lstrip('/'))
        logger.info('Generating section report for %s' % secname) 

        report_params = {
            'secname':secname, 
            'inpath':latex_string(self.inpath),
            'outpath':latex_string(self.outpath),        
            'plot_dim':self._plot_dim,
            'parameters':self._run_summary,
            # Hacky way to look for plots (leave it to the tex template) to use 
            'specplotfull': op.join(self.outpath, 'Full_data/Raw_spectrum'),
            'areaplotfull': op.join(self.outpath, 'Full_data/Raw_area'),
            
            'specplotabs': op.join(self.outpath,  'Abs_base10/Absorbance_spectrum'),
            'areaplotabs': op.join(self.outpath,  'Abs_base10/Absorbance_area'),
            
            'specplotrel':op.join(self.outpath, 'Linear_ratio/Relative_spectrum'),
            'areaplotrel':op.join(self.outpath, 'Linear_ratio/Relative_area')
                        } 

        report_path = op.join(self.outpath, 'sectionreport.tex')
        report = open(report_path, 'w')
        report.write( SECTION % report_params )                        
        report.close()
        
        logger.debug("Adding %s to tree dic." % self.infolder )
        self._treedic[latex_string(secname)] = report_path


    def _analyze_dir(self):
        """ Analyze a single directory, creates timespectra; performs analysis.
            Very much the main method in the whole class."""

        logger.info("ANALYZING RUN DIRECTORY: %s\n\n" % self.infolder)

        # outpath should be correctly set by main_walk()
        rundir = self.outpath
        logmkdir(rundir)

        start = timenow()
        ts_full = self.build_timespectra()        
        logger.info('SUCCESS imported %s in %s seconds' % 
                (ts_full.full_name, (timenow() - start).seconds))
        ts_full = self.validate_ts(ts_full) 
        
        logger.info('Saving %s as %s.pickle' % (ts_full.full_name, self.infolder))
        ts_full.save(op.join(rundir, '%s.pickle' % self.infolder)) 
        
        # Output metadata to file (read back into _run_summary)
        if getattr(ts_full, 'metadata', None):
            logger.info('Saving %s metadata' % ts_full.full_name)
            with open(op.join(rundir, '%s.full_metadata' % self.infolder), 'w') as f:
                f.write(dict_out('Spectrometer Parameters', ts_full.metadata))

            def _filter_metadata(dic):
                """ Return spectrometer parameters of interest for report. Changes
                    format of integration time parameter."""
                dic = OrderedDict( (k, (dic.get(k, '') )) for k in ['int_time', 'int_unit', 
                         'boxcar', 'spec_avg', 'timestart', 'timeend', 'filecount','spectrometer'])

                if 'usec' in dic['int_unit']:
                    dic['int_unit'] = 'usec'

                dic['int_time'] = str(dic['int_time']) + ' ' +  str(dic['int_unit'])
                del dic['int_unit']
                return dic
                
            self._run_summary = latex_multicols(_filter_metadata(ts_full.metadata), title='Spectrometer Parameters')

        else:
            logger.info('Metadata not found for %s' % ts_full.full_name)
            

        # To csv (loses metadata) set to meta_separate to False to preserve metadata
        logger.info('Outputting to %s.csv.  Metadata will be exluded.' % self.infolder)
        
        self.save_csv(ts_full, op.join(rundir, '%s.csv' % self.infolder))

        # Set ts, subtract baseline, crop
        ts = self.apply_parameters(ts_full)     # BASELINE SUBTRACTED/CROPPING HERE   
         
        cropped_csv_path =  op.join(rundir, '%s_cropped.csv' % self.infolder)
        self.save_csv(ts, cropped_csv_path)
        
        # Used to be in apply parameters, but pulled out so could save bline_cropped csv
        # Once jsonify, this can be returned to apply parameters fcn 
        if self.params.intvlunit:
            try:
                ts.varunit = self.params.intvlunit
            except KeyError:
                ts.varunit = 'intvl'        
                logger.warn('Cannot set "intvlunit" from parameters; running'
                        ' ts.to_interval()')
        else:
            logger.info('Intvlunit is None- leaving data as rawtime')
        
        #IPYTHON NOTEBOOK
        NBPATH =  op.join(rundir, '%s.ipynb' % self.infolder)
        logging.info("Copying blank .ipynb template to %s" % NBPATH)
        template = open(IPYNB, 'r').read()
        
        outshort = self.outpath.split(self.outroot)[1].lstrip('/')

        template = template.replace('---DIRECTORY---', outshort)
        template = template.replace('---CREATED---',  strftime("%Y-%m-%d %H:%M:%S", gmtime()))
        template = template.replace('---ROOT---', hideuser(self.outroot))
        template = template.replace('---PARAMS---', self.params.as_markdownlist())
        template = template.replace('---COUNT---', str(ts.shape[1]))
        template = template.replace('---HIDDENOUT---', hideuser(rundir))

        cols = ts.shape[1]
        _warncolor = 'green'
        if cols >= 200 and cols < 500:
            _warncolor = 'orange'
        elif cols >= 500:
            _warncolor = 'red'
        template = template.replace('---COLOR---', _warncolor)
        
        
        # Pass the full csv file otherwise run_nb_offline won't be in right wd
        template = template.replace('---CSVPATH---', op.basename(cropped_csv_path))
        
        template = template.replace('---TSTART---', '%s'%ts_full.columns[0])
        template = template.replace('---TEND---', '%s'%ts_full.columns[-1])
        template = template.replace('---TDELTA---', '%s'%
                                    ts_full.as_varunit('intvl').columns[-1])                            

        open(NBPATH, 'w').write(template)

        # ADD NOTBOOK TO GITHUB
        if self.params.git:
            os.system("git add %s" % op.abspath(NBPATH))
            viewerpath = NBPATH.split('FiberData')[1].lstrip('/')
            viewerpath = 'http://nbviewer.ipython.org/github/hugadams/FiberData/blob/master/'+viewerpath
            localdir = self.inpath.replace(self.inroot, '').lstrip('/')
            NBVIEWPATHS.append((localdir, viewerpath)) 
       
        
        # Execute the notebook
#	        run_nb_offline(NBPATH) 
        #try:
            #from runipy.notebook_runner import NotebookRunner
            #from IPython.nbformat.current import read as nbread
            #from IPython.nbformat.current import write as nbwrite
        #except IOError:
            #logger.critical("Please install runipy (pip install runipy) to "
                            #"run .ipynb prior to opening.")
        #else:
            #notebook = nbread(open(NBPATH), 'json')
            #r = NotebookRunner(notebook)
            #r.run_notebook()            
            #nbwrite(r.nb, open(NBPATH, 'w'), 'json')
    
        # Quad plot (title is rootfolder:folder; for non -s, are the same
        if op.basename(self.inroot) == self.infolder:
            quadname = self.infolder 
        else:
            quadname = "%s:%s" % (op.basename(self.inroot), self.infolder)

        # SIXPLOT
        six_plot(ts, title=quadname, striplegend=True)
        self.plt_clrsave(op.join(rundir, '%s_sixplot.png' % self.infolder))
        
        # AREA THIRDS
        self.area_thirds_plot(ts)
        self.plt_clrsave(op.join(rundir, '%s_area_thirds.png' % self.infolder))
        
        
        #Iterate over various norms
        for iu in self.params.norms:
            od = op.join(rundir , _normdic[iu])
            # Rename a few output units for clear directory names
            if iu =='r':
                od = op.join(rundir, 'Linear_ratio') 
            elif iu == None:
                od = op.join(rundir, 'Full_data')
            elif iu == 'a':
                od = op.join(rundir, 'Abs_base10')
            logmkdir(od)
             
            ts = ts.as_norm(iu)
            out_tag = ts.full_norm.split()[0] #Raw, abs etc.. added to outfile
            
            
            # 1d Plots
            if '1d' in self.analysis:
                self.plots_1d(ts, outpath=od, prefix=out_tag)
            
            # 2d Plots
            if '2d' in self.analysis:
                self.plots_2d(ts, outpath=od, prefix=out_tag)
            
            # Correlation analysis  
            if 'corr' in self.analysis:
                self.corr_analysis(ts, outpath=od, prefix=out_tag)
          
            # 3d Plots
            if '3d' in self.analysis:
                self.plots_3d(ts, outpath=od, prefix=out_tag)
                          
            

    def apply_parameters(self, ts):
        """ Performs several timespectr manipulations such as slicing, baseline 
            subtraction etc..  similar to self.validate_ts(), but is performed
            after the full timespectra object has been changed. """
        

        ## Subtract the dark spectrum if it has one.  
        #if self.params.sub_base:
            #if ts.baseline is None:
                #logger.warn('Warning: baseline not found on %s)' % ts.full_name)            
            #else:    
                #ts.sub_base() 
                #logger.info('Baseline has been subtracted')
        #else:
            #logger.info('Parameter "sub_base" not set to True.' 
            #'No baseline subtraction will occur.')            
    
        # Fit first order references automatically?
        
        if self.params.bline_fit:
            blines = dynamic_baseline(ts, self.params.fit_regions )
            ts = ts.sub(blines, axis=0)
            logger.info('Dynamically fit baselines successfully subtracted.')
    
        # Slice spectra and time start/end points.  Doesn't stop script upon erroringa
        try:
            tstart, tend = self.params.t_start, self.params.t_end
            if tstart and tend:
                ts = ts[ tstart : tend ] 
                logger.info('Timesliced %s to (%s, %s)' % (ts.full_name, tstart, tend))             
            else:
                logger.debug('Cannot timeslice: tstart and/or tend not found')
        except Exception:
            logger.warn('unable to timeslice to specified by parameters')                   
            raise
            
        try:
            xstart, xend = self.params.x_min, self.params.x_max
            if xstart and xend:              
                ts = ts.ix[ xstart : xend ] 
                logger.info('Wavelength-sliced %s to (%s, %s)' % (ts.full_name, xstart, xend))                
            else:
                logger.debug('Cannot spectralslice: xstart and/or xend not found')

        except Exception:
            logger.warn('unable to slice x_min and x_max range specified by parameters')                          
            
            
        # Set reference (MUST SET reference AFTER SLICING RANGES TO AVOID ERRONEOUS reference DATA)
        ts.reference = self.params.reference
                        
        return ts    

    def build_outroot(self):
        if self.outroot == self.inroot:
            raise ParameterError('Inroot and outroot cannot be the same! '
                'Input data would be lost...')
        if op.exists(self.outroot):
            if not self.overwrite:
                raise IOError("Outdirectory already exists!")                
            else:
                logger.warn('Directory "%s" and its contents will be overidden' % 
                            op.basename(self.outroot))
                shutil.rmtree(self.outroot)
        logmkdir(self.outroot)

        
    def validate_ts(self, ts):
        """ Tries to catch errors in index, as well as sets some defaults.  """ 
        
        logger.info('Testing spectral index bounds against parameters min/max')

        pmin, pmax = self.params.valid_minmax                
        if  ts.index[0] < pmin or ts.index[-1] > pmax:
            raise GeneralError('Parameters _valid_range criteria not met.  Data' 
            ' range (%s,%s) - parameters "_valid_range" (%s,%s)' % 
            (ts.index[0], ts.index[-1], pmin,pmax))  
            
        ts.specunit = self.params.specunit 
        ts.intvlunit = self.params.intvlunit 
        
        # TEST FOR NAN VALUES BEFORE CONTINUING  #####
        if countNaN(ts): #0 will evaluate to false
            raise GeneralError('NaNs found in %s during preprocessing.' 
                    ' Cannot proceed with analysis.' % ts.full_name)
        return ts


    def build_timespectra(self):
        """ Attempts to build timespectra form picklefile, legacy and finally
            from the raw datafiles."""
                
        # Files in working directory, ignore certain extensions; ignore directories
        infiles = get_files_in_dir(self.inpath, sort=True)
        infiles = [f for f in infiles if not ext(f) in self.img_ignore]
                
        if not infiles:
            raise IOError("No valid files found in %s" % self.infolder)        
        
        # Try get timespectra from picklefiles
        ts_full = self._ts_from_picklefiles(infiles)
        if ts_full:
            return ts_full

        # Try get timespectra from timefile/datafile (legacy)
        ts_full = self._ts_from_legacy(infiles)
        if ts_full:
            return ts_full
        
        # Import from raw specfiles (overwrite if data < 1s apart)
        logger.info('Loading contents of %s multiple raw spectral '
                    'files %s.' % (self.infolder, len(infiles)))     
        try:
            return from_spec_files(infiles, name=self.infolder, 
                                   check_for_overlapping_time=False) 
        except Exception as exc:
            logger.critical('Could not import files from pickle, legacy or' 
            ' from_spec_files()')
            raise
        
        
    def _ts_from_legacy(self, infiles):
        
        if len(infiles) != 2:
            logger.debug('Legacy import failed: requires 2 infiles (timefile/darkfile); '
            '%s files found' % len(infiles))
            return

        timefile=[afile for afile in infiles if 'time' in afile.lower()]            

        if len(timefile) != 1:
            logger.debug('Legacy import failed: required one file with "time" in'
            ' filenames; found %s' % len(infiles))
            return
        
        logger.info('Loading contents of "%s", using the timefile/datafile '
                    'conventional import' % self.infolder)                       
        timefile = timefile[0]
        infiles.remove(timefile) #remaing file is datafile
        return from_timefile_datafile(datafile=infiles[0], timefile=timefile, 
                                      name=self.infolder)

    @classmethod
    def _ts_from_picklefiles(cls, infiles, infolder='unknown'):
        """ Look in a list of files for .pickle extensions.  Infolder name
            is only used for better logging. """

        logger.debug('Looking for .pickle files in folder: %s' % infolder)
        
        picklefiles = [f for f in infiles if ext(f) == 'pickle']
        
        if not picklefiles:
            logger.debug('Pickle files not found in %s' % infolder)
            return
        
        elif len(picklefiles) > 1:
            raise IOError('%s pickle files found; expected only 1')
        
        else:
            picklefile = picklefiles[0]
            logger.info('Loaded contents of folder, %s, using the ".pickle" ' 
                       'file, %s.' % (folder, op.basename(picklefile)))            
            return mload(picklefile)     
 
    def plots_1d(self, ts, outpath, prefix=''):
        """ Plots several 1D plots.  User passes in ts w/ proper norm.
            outpath: filepath (str)
            prefix: str
               Put in front of file name (eg outpath/prefix_area.png) for area plot
        """
        
        # Set plot and tick size larger than defaul
        sizeargs = {'labelsize': self._plot_fontsize, 
                    'ticksize':15, 
                    'titlesize':self._plot_fontsize}
        
    
        ts.plot(**sizeargs)
        self.plt_clrsave(op.join(outpath, prefix +'_spectrum'))

        # Area plot using simpson method of integration
        areaplot(ts, 
                 ylabel='Power',
                 xlabel='Time ('+ts.varunit+')',
                 legend=False,
                 title='Spectral Power vs. Time (%i - %i %s)' % 
                    (min(ts.index), max(ts.index), ts.specunit), color='black', **sizeargs)
        self.plt_clrsave(op.join(outpath, prefix +'_area'))

        # Short wavelengths min:525nm
        areaplot(ts.loc[:525.0, :], 
                 ylabel='Power',
                 xlabel='Time ('+ts.varunit+')',
                 legend=False,
                 title='Short Wavelengths vs. Time (%i - %i %s)' % 
                    (min(ts.index), 525.0, ts.specunit), color='b', **sizeargs)
        self.plt_clrsave(op.join(outpath, prefix +'_area_short'))

        # Middle wavelengths 525:590nm
        areaplot(ts.loc[525.0:590.0, :], 
                 ylabel='Power',
                 xlabel='Time ('+ts.varunit+')',
                 legend=False,
                 title='Medium Wavelengths vs. Time (%i - %i %s)' % 
                    (525.0, 590.0, ts.specunit), color='g', **sizeargs)
        self.plt_clrsave(op.join(outpath, prefix +'_area_middle'))
        
        # Long wavelenghts 590.0nm:
        areaplot(ts.loc[590.0:, :], 
                 ylabel='Power',
                 xlabel='Time ('+ts.varunit+')',
                 legend=False,
                 title='Long Wavelengths vs. Time (%i - %i %s)' % 
                    (590.0, max(ts.index), ts.specunit), color='r', **sizeargs)
        self.plt_clrsave(op.join(outpath, prefix +'_area_long'))        

        # Normalized area plot (divided by number x units)       
        areaplot(ts/len(ts.index),
                 ylabel='Power per unit %s' % ts.specunit,
                 xlabel='Time (' + ts.varunit+')', 
                 legend=False, 
                 title='Normalized Spectral Power vs. Time (%i - %i %s)' % 
                     (min(ts.index), max(ts.index), ts.specunit), color='orange', **sizeargs)
        self.plt_clrsave(op.join(outpath, prefix +'_area_normal'))
        
        # Normalized ABSOLUTE areaplot (divided by number x units)       
        areaplot(np.absolute(ts)/len(ts.index), 
                 ylabel='Power per unit %s' % ts.specunit, 
                 xlabel='Time (' + ts.varunit+')', 
                 legend=False, 
                 title='Normalized ABSOLUTE Power vs. Time (%i - %i %s)' % 
                     (min(ts.index), max(ts.index), ts.specunit), color='purple', **sizeargs)
        self.plt_clrsave(op.join(outpath, prefix +'_area_normal'))   
        
        
        
        
        # Lambda max vs. t plot
        
        # XXXXXX?

        # Ranged time plot
        try:
            uv_ranges = self.params.uv_ranges
        except (KeyError):
            uv_ranges = 8   
            logger.warn('Uv_ranges parameter not found: setting to 8.')
                
        # Time averaged plot, not scaled to 1 (relative intenisty dependson bin width and actual intensity)
        tssliced = ts.wavelength_slices(uv_ranges, apply_fcn='mean')
        range_timeplot(tssliced, ylabel='Average Intensity', xlabel = 
                       'Time ('+ts.varunit+')', **sizeargs ) #legstyle =1 for upper left
        self.plt_clrsave(op.join(outpath, prefix +'_strip'))        
        
    def area_thirds_plot(self, ts):
        # Special areaplot of normalized raw area in thirds
        # Should work on USB650, but untested.  Slices data into thirds to 
        # find limits, so should be independent.
        def _scaleto1(tslice):
            return np.divide(tslice, tslice[0])
        
        ax = areaplot(_scaleto1(ts.area()), linewidth=2, alpha=1, ls='--', custompadding=None)
        
        tspace = (ts.index.max() - ts.index.min()) /3
        tim = ts.index.min()
        
        short = ts.nearby[:tim+tspace].area()
        mid = ts.nearby[tim+tspace:tim+2*tspace].area()
        longer = ts.nearby[2*tspace+tim:].area()
        
        # Store slice ranges for plt.legend() below
        label_short = '%s:%s'% (ts.index[0], tim+tspace)
        label_mid = '%s:%s' % (tim+tspace, tim+2*tspace)
        label_long = '%s:%s' % (2*tspace+tim, ts.index[-1])
        
        for t, color in [(short, 'b'), (mid, 'g'), (longer, 'r')]:
            ax = range_timeplot(_scaleto1(t), ax=ax, color=color, 
                                linewidth=4, alpha=.4, custompadding=None)
        
        plt.title('Area slices %s: normalized at t=0' % ts.specunit)    
        plt.legend(['All $\lambda$', label_short, label_mid, label_long],
                    ncol=4, 
                    fontsize='10')        

        
    def corr_analysis(self, ts, outpath, prefix=''):
        
        # Make a 2dCorrAnal Directory
        corr_out = op.join(outpath, '2dCorrAnal')
        logmkdir(corr_out)
    
        ref = make_ref(ts, method='empty') 
        S,A = ca2d(ts, ref)
        sync_3d(S, title='Synchronous Spectrum (%s-%s %s)' % (round(min(ts), 1),
                round(max(ts),1), ts.full_varunit)) #min/max by columns      
        self.plt_clrsave(op.join(corr_out, prefix +'_sync'))        
        async_3d(A, title='Asynchronous Spectrum (%s-%s %s)' % 
                 (round(min(ts), 1), round(max(ts),1), ts.full_varunit))
        self.plt_clrsave(op.join(corr_out, prefix +'_async'))        

        
    def plots_2d(self, ts, outpath, prefix=''):
        
        plot2d(ts, title='Full Contour', cmap='autumn', contours=7, label=1,
               colorbar=1, background=1)
        self.plt_clrsave(op.join(outpath, prefix +'_contour'))    
        
        
    def plots_3d(self, ts, outpath, prefix=''):
        c_iso = 10 ; r_iso=10
        kinds = ['contourf', 'contour']
        views = ( (14, -21), (28, 17), (5, -13), (48, -14), (14,-155) )  #elev, aziumuth
        for kind in kinds:
            out3d = op.join(outpath, '3dplots_'+kind)
            logmkdir(out3d)

            for view in views:        
                spec_surface3d(ts, kind=kind, c_iso=c_iso, r_iso=r_iso, 
                    cmap=cmget('gray'), contour_cmap=cmget('autumn'),
                    elev=view[0], azim=view[1], xlabel='Time ('+ts.varunit+')'
                              )
                self.plt_clrsave(op.join(out3d, prefix + 'elev:azi_%s:%s' %
                                         (view[0], view[1])))         
   
    def plt_clrsave(self, outpath):
        logger.info('Saving plot: %s' % outpath)                
        plt.savefig(outpath, dpi=self._plot_dpi)
        plt.clf()       

    @classmethod
    def from_namespace(cls, args=None):
        """ Create Controller from argparse-generated namespace. """
        
        if args:
            if isinstance(args, basestring):
                args=shlex.split(args)
            sys.argv = args               
        
        
        parser = argparse.ArgumentParser(SCRIPTNAME, description='GWU skspec fiber data '
        'analysis.', epilog='Additional help not found', 
        usage='%s <indir> <outdir> --globals' % SCRIPTNAME)

        # Global options
        parser.add_argument('inroot', metavar='indir', action='store', default=DEF_INROOT, 
                          help='Path to root directory where file FOLDERS are '
                          'located.  Defaults to %s' % DEF_INROOT)
        
        parser.add_argument('outroot', metavar='outdir', action='store', default = DEF_OUTROOT,  
                          help = 'Path to root output directory.  Defaults to %s'
                          % DEF_OUTROOT) 
        
        parser.add_argument('-s','--sweep', action='store_true', 
                            help='Script will recursively analyze directories '
                            'from the top down of indir')

        parser.add_argument('-g', '--git', action='store_false', help='If true, '
                            'git will not attempt to be called on this run.'
                            'By default, git is synced.')
    
        parser.add_argument('-c', '--config', dest='cfig', metavar='', 
                           choices=['usb650', 'usb2000'], default='usb2000',
                           help='usb650, usb2000, defaults to usb2000')            
         
        parser.add_argument('-o','--overwrite', action='store_true', 
                            help='Overwrite contents of output directories if '
                            'they already exist')
        
        parser.add_argument('-v', '--verbosity', help='Set screen logging '
                             'If no argument, defaults to info.', nargs='?',
                             default='warning', const='info', metavar='')

                
        parser.add_argument('-p','--params', nargs='*', help='Overwrite config '
                            'parameters manually in form k="foo value" '  
                            'Ex: --params x_min=440.0 x_max=700.0', metavar='')
        
        parser.add_argument('-a', '--analysis', nargs='*', help='Types of '
                'analysis to perform.  Choose 1 or more of the following: %s or '
                '"all".  Defaults to %s.' % (ALL_ANAL, ANAL_DEFAULT), 
                default=ANAL_DEFAULT, metavar='')
        
        parser.add_argument('-e', '--extra', action='store_true', 
            help='Adds report and sem directories to current working directory.'
            ' Not modular and only a convienence hack')
        
        parser.add_argument('-f', '--fontsize', type=int, default=18,
             help='Plot x/y/title fontsize.  Defaults to 18.  pyplot default = 12.')
    
#        parser.add_argument('-d','--dryrun', dest='dry', action='store_true',
#                            help='Not yet implemented')

        # This must be "-t, --trace" for logger compatability; don't change!
        parser.add_argument('-t', '--trace', action='store_true', dest='trace',
                          help='Show traceback upon errors')       
        
        parser.add_argument('--plot_dim', metavar='', default='width=6cm', help='Defaults to '
            '"width=6cm"; any valid latex plotsize parameters (\textwidth) are acceptable;' 
            ' enter directly as they would be in latex "includegraphics[]".')
        parser.add_argument('--dpi', type=int, help='Plotting dots per inch.')
    
    
        # Store namespace, parser, runn additional parsing
        ns = parser.parse_args()

        # Run additional parsing based on cfig and "params" to set spec params
        if ns.cfig == 'usb650':
            model = USB650
    
        elif ns.cfig == 'usb2000':
            model = USB2000
    
        if not ns.params:
            ns.params=[]
        
        try:
            userparams = dict (zip( [x.split('=', 1 )[0] for x in ns.params],
                                [x.split('=', 1)[1] for x in ns.params] ) )
    
        except Exception:
            raise IOError('Please enter keyword args in form: x=y')
        
        userparams['git'] = ns.git
        ns.params = model(**userparams)
        delattr(ns, 'cfig')        
        
        controller = cls(inroot=ns.inroot, outroot=ns.outroot, plot_dpi = ns.dpi,
                   verbosity=ns.verbosity, trace=ns.trace, params=ns.params, 
                   overwrite=ns.overwrite, sweep=ns.sweep, plot_dim = ns.plot_dim,
                   analysis=ns.analysis, fontsize=ns.fontsize)
            
        # Make REPORT/SEM directories for convienence
        if ns.extra:
            for dirname in ['Report', 'SEM']:
                logger.info('Making directory: "%s"' % dirname)
                if op.exists(dirname):
                    logger.critical('Cannot create directory: %s.  Not '
                        'comfortable overwriting...' % dirname)
                else:
                    os.mkdir(dirname)
                if dirname == 'Report' and not op.exists(dirname):
                    os.mkdir(op.join(dirname, 'images'))
                    
        return controller



''' Serialization interface for custom DataFrame objects.  Allows to save/load 
 for memory streams or files.  Because one cannot serialize DataFrames with
 custom attributes, this uses an intermediate object for that process.  Plan
 it implement pickling saved methods later (requires more work).  These are meant to 
 supplant the DataFrame's save() and load() methods when custom attributes must persist.
 
 Note, this program assesses custom attributes by inspecting your DataFrame's
 attributes using Python's builting function, dir().  It compares these to the
 attributes of an empty DataFrame.  This adds a bit of overhead, but should allow
 this program to work with new versions of pandas, as Dataframe's methods and attributes
 are likely to change.  Is there are better way to do this?
 
 The following four functions are defined:
   df_dumps:  Serialize a DataFrame into memory.  Returns serialized stream.
   df_dump:  Serialize a DataFrame into a file.  Returns None.
   df_loads: Return a DataFrame from a serialized stream.
   df_load: Return a Dataframe from a serialized file.
 
 See bottom of file for test cases:  '''

__author__ = "Adam Hughes"
__maintainer__ = "Adam Hughes"
__email__ = "hugadams@gwmail.gwu.edu"
__status__ = "Prototype"

import cPickle
from operator import attrgetter
from pandas import DataFrame

### For testing ###
from numpy.random import randn


class TempDump(object):
    ''' Temporary class to dump DataFrame object with custom attributes.  Custom attrubutes are
    passed in as a dictionary and then temporarily stored upon serialization as _metadict.  Upon
    deserialization, the attributes and values are re-appended to the DataFrame automatically.'''
    def __init__(self, dataframe, metadict):
        self.dataframe=dataframe
        self._metadict=metadict

dfempty=DataFrame() 
defattrs=dir(dfempty)

def print_customattr(df):
    '''Formatted output of all custom attributes found in a DataFrame.  For all
    attributes and methods, use dir(df).'''
    metadict=_get_metadict(df)
    if len(metadict) > 0:
        print '\nFound %s custom attributes:\n'%len(metadict)        
        print '\n'.join([(k+'\t'+v) for k,v in sorted(metadict.items())])

    else:
        print 'No custom attributes found'

def _get_metadict(df):
    ''' Returns dictionary of attributes in a dataframe not found in the default frame.'''
    attrs=dir(df)
    newattr=[att for att in attrs if att not in defattrs] #if not is type(instancemethod?)
    if len(newattr) > 1:
        fget=attrgetter(*newattr)
        return dict(zip(newattr, fget(df)))
    else:
        return {}

def df_dumps(df):
    ''' Save dataframe as a stream into memory.'''
    metadict=_get_metadict(df)
    return cPickle.dumps(TempDump(df, metadict )) #Dumps writes the object to memory
    
def df_dump(df, outfile):
    ''' Save dataframe as a file.'''
    outstream=df_dumps(df) #Dumps writes the object to memory    
    f=open(outfile, 'w') #Should this be 'wb'
    f.write(outstream)
    f.close()
    return None #Should I return none or stream?    
    
def df_load(infile):
    '''Returns dataframe from a serialized file '''
    f=open(infile, 'r')
    tempobj=cPickle.load(f)
    f.close()
    df=tempobj.dataframe    
    for attr, value in tempobj._metadict.items():
        setattr(df, attr, value)
    return df       

def df_loads(stream):
    ''' Returns dataframe from a serialized stream'''
    tempobj=cPickle.loads(stream) #loads not load
    df=tempobj.dataframe
    for attr, value in tempobj._metadict.items():
        setattr(df, attr, value)
    return df    

    
if __name__ == '__main__':
    ### Make a random dataframe, add some attributes
    df=DataFrame(((randn(3,3))), columns=['a','b','c'])
    print_customattr(df)
    print 'adding some attributes'
    df.name='Billy'
    df.junk='in the trunk'
    print_customattr(df)
    
    ### Serialize into memory
    stream=df_dumps(df)
    print 'wrote dataframe to memory'
    ### Restore from memory
    dfnew=df_loads(stream)
    print 'restored from memory'
    print_customattr(dfnew)
    
    
    ### Serialize into file
    outfile='dftest.df' #What file extension is commonly used for this?
    df_dump(df, outfile)  
    print 'wrote dataframe to file %s'%outfile
    ### Restore from file     
    dfnewnew=df_load(outfile)
    print 'Restored from file%s'%outfile
    print_customattr(dfnewnew)





''' Various dataframe-compatible reference utilities. '''

__author__ = "Adam Hughes"
__copyright__ = "Copyright 2012, GWU Physics"
__license__ = "Free BSD"
__maintainer__ = "Adam Hughes"
__email__ = "hugadams@gwmail.gwu.edu"
__status__ = "Development"

import numpy as np
from pandas import DataFrame, Series
import logging

logger = logging.getLogger(__name__)

        
def _find_nearest(array, value):
    '''Find nearest value in an array, return index and array value'''
    idx=(np.abs(array-value)).argmin()
    return idx, array[idx]      

def dynamic_baseline(df, slices, style='linear', weightstyle=None, axis=1):
    '''Applies a dynamically calculated reference correction to a dataframe.  User passes in index values, either ranges
    or points, then a linear fit is applied through these points.  Each column in the dataframe is scaled its corresponding
    curve, which is then subracted off.  Essentially, this subracts a reference that can be lopsided- useful for assymetric data, and
    is usually how crude ATR-IR reference correction is performed.  
    
    For now, program only works for index values as the reference.
    
    df- DataFrame.
    
    slice- List of points or ranges at which the reference data is sampled.  For example, 
           ( (300.0, 500.0), 900.0 ).  Curve will take all index values of the dataframe between
           300.0:500.0 (nanometers in my case), and also take the closest index value there is to 900.0. 
           A fit is then drawn between these points.
           
    style- Fitting style to connect the dots between slice regions.  For now, only linear fit is used.  Not sure if higher order
           will ever be useful, so I left this in here as a reminder.
           
    axis-  Axis overwhich to do this reference correction.  For now, ONLY INDEX IS IMPLEMENTED.
    
    weightstyle- Tries to account for idea that if I pass slices, and one slice contains more points, then the line of best fit will be weighted
                 more strongly to this region.  For now, I haven't implemented this, but would have to actually use poly1d to call more points to the
                 other regions, or some other solution?  This does have an option, mean, which will take the mean if each range so that each range gets 
                 a single fit point; however, I think this is useless as this can be passed in anyway.  And it's the mean of the x-values, not even the 
                 mean of the data along that axis anyway.
                 
    returns-
       DataFrame of fitted references.'''
        
    ### Test for proper input ###
    if style != 'linear' or axis != 1:
        raise NotImplementedError('reference correction only support linear fitting style and index axis iteration only.')
    
    if weightstyle:
        if weightstyle != 'midpoint':
            raise NotImplemented('weightstyle attribute in reference_correction must be either None or "midpoint" but %s was entered'%weightstyle)

    xp=[] #xpoints
    
    logger.info('Applying dynamic baseline (style=%s)' % style)
    
    for val in slices:
        ### If slice is a single point, find its nearest actual entry in df.index
        if len(val) == 1:
            xp.append(_find_nearest(df.index, val)[1])
        
        ### If slice is range, take all index values in that region
        elif len(val) == 2:
            xvals=np.asarray(df.ix[ val[0]: val[1] ].index)

            if weightstyle =='mean':
                xp.append(np.mean(xvals))

            else:
                xp=xp+list(xvals) #List necessary to get npdtype straight..
            
            
        else:
            raise AttributeError('In reference correction, slices must 1 or 2 items;'
                'received %s of len %s'%(val, len(val)))
    
    ### Apply 1-d connect the dots curvefit to each column, then subtract this from said column
    ### Still vectorized, despite loop.  Loop is necessary; too much to use df.apply.
    bout={}                      
    for col in df.columns:
        curve=df[col] #Series at this point
        z=np.polyfit(xp, curve[xp], 1)  
        p=np.poly1d(z)        
        array=p(np.asarray(list(curve.index) ) )   
        reference=Series(array, name=col, index=curve.index) #curve.index instead of df.index                            
        bout[col]=reference #incase index is getting recast for nans or some behavior I may not be aware of
    
    bout=DataFrame(bout)
    bout.reindex(columns=df.columns) #Columns are automatically if they are datetimes
    return bout        


from dataframeserial import _get_metadict

### Following two functions deal with attribute stored of dataframe.
def store_attr(df):
    ''' Store all the attributes in a dataframe that are not found in an empty dataframe
    declared with DataFrame().'''
    return _get_metadict(df)
    
def restore_attr(df, stored_attr):
    ''' Set the attributes of the pandas dataframe using sotred_attr dictionary.'''
    for attr, value in stored_attr.items():
        setattr(df, attr, value)    
    return df

def transfer_attr(df1, df2, reversedeletion=False, speakup=True):
    ''' Transfer all the attributes from df1 to df2.  If none found, returns df2.
    Speakup keyword just adds print statements.
    Reverse deletion means that all attributes in df2 that are not found in df1.'''
    if reversedeletion:
        raise NotImplementedError('Didnt put that option in yet chief.')
    
    df1attr=store_attr(df1)

    if len(df1attr) > 0:
        df2=restore_attr(df2, df1attr)
        if speakup:            
            print '\nTransferring %s attributes:\n'%len(df1attr)        

    else:
        if speakup:
            print 'No attributes found for transfer'
    
    return df2
     

